##############################################################################
#
# Copyright (c) 2003, 2018 Zope Foundation and Contributors.
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################
"""ZConfig factory datatypes for log handlers."""

import functools
import sys
import urllib.parse
from abc import abstractmethod

import ZConfig.components.logger.formatter
from ZConfig.components.logger.factory import Factory


_log_format_variables = (
    ZConfig.components.logger.formatter._log_format_variables)


def log_format(value):
    value = ZConfig.components.logger.formatter.ctrl_char_insert(value)
    try:
        # Make sure the format string uses only names that will be
        # provided, and has reasonable type flags for each, and does
        # not expect positional args.
        value % _log_format_variables
    except (ValueError, KeyError):
        raise ValueError('Invalid log format string %s' % value)
    return value


# Backward compatibility, because paranoia is our friend:
ctrl_char_insert = ZConfig.components.logger.formatter.ctrl_char_insert
resolve = ZConfig.components.logger.formatter.resolve


class HandlerFactory(Factory):

    def __init__(self, section):
        Factory.__init__(self)
        self.section = section
        factory = ZConfig.components.logger.formatter.FormatterFactory(section)
        self.create_formatter = factory

    @abstractmethod
    def create_loghandler(self):
        "subclasses must override create_loghandler()"

    def create(self):
        logger = self.create_loghandler()
        logger.setFormatter(self.create_formatter())
        logger.setLevel(self.section.level)
        return logger

    def getLevel(self):  # pragma: no cover Is this used?
        return self.section.level


class FileHandlerFactory(HandlerFactory):

    def __init__(self, section):
        HandlerFactory.__init__(self, section)

        from ZConfig.components.logger import loghandler
        path = section.path
        max_bytes = section.max_size
        old_files = section.old_files
        when = section.when
        interval = section.interval
        encoding = section.encoding
        delay = section.delay

        def check_std_stream():
            if max_bytes or old_files or when:
                raise ValueError("cannot rotate " + path)
            if delay:
                raise ValueError("cannot delay opening " + path)
            if encoding:
                raise ValueError("cannot specify encoding for " + path)

        if path == "STDERR":
            check_std_stream()

            def factory():
                return loghandler.StreamHandler(sys.stderr)

        elif path == "STDOUT":
            check_std_stream()

            def factory():
                return loghandler.StreamHandler(sys.stdout)

        elif when or max_bytes or old_files or interval:
            if not old_files:
                raise ValueError("old-files must be set for log rotation")
            if when:
                if max_bytes:
                    raise ValueError("can't set *both* max_bytes and when")
                if not interval:
                    interval = 1
                factory = functools.partial(
                    loghandler.TimedRotatingFileHandler,
                    path, when=when, interval=interval,
                    backupCount=old_files, encoding=encoding, delay=delay)
            elif max_bytes:
                factory = functools.partial(
                    loghandler.RotatingFileHandler,
                    path, maxBytes=max_bytes, backupCount=old_files,
                    encoding=encoding, delay=delay)
            else:
                raise ValueError(
                    "max-bytes or when must be set for log rotation")
        else:
            factory = functools.partial(
                loghandler.FileHandler,
                path, encoding=encoding, delay=delay)

        self._factory = factory

    def create_loghandler(self):
        return self._factory()


_syslog_facilities = {
    "auth": 1,
    "authpriv": 1,
    "cron": 1,
    "daemon": 1,
    "kern": 1,
    "lpr": 1,
    "mail": 1,
    "news": 1,
    "security": 1,
    "syslog": 1,
    "user": 1,
    "uucp": 1,
    "local0": 1,
    "local1": 1,
    "local2": 1,
    "local3": 1,
    "local4": 1,
    "local5": 1,
    "local6": 1,
    "local7": 1,
    }


def syslog_facility(value):
    value = value.lower()
    if value not in _syslog_facilities:
        L = sorted(_syslog_facilities.keys())
        raise ValueError("Syslog facility must be one of " + ", ".join(L))
    return value


class SyslogHandlerFactory(HandlerFactory):

    def create_loghandler(self):
        from ZConfig.components.logger import loghandler
        return loghandler.SysLogHandler(self.section.address.address,
                                        self.section.facility)


class Win32EventLogFactory(HandlerFactory):

    def create_loghandler(self):
        from ZConfig.components.logger import loghandler
        return loghandler.Win32EventLogHandler(self.section.appname)


def http_handler_url(value):
    scheme, netloc, path, param, query, fragment = urllib.parse.urlparse(value)
    if scheme != 'http':
        raise ValueError('url must be an http url')
    if not netloc:
        raise ValueError('url must specify a location')
    if not path:
        raise ValueError('url must specify a path')
    q = []
    if param:
        q.append(';')
        q.append(param)
    if query:
        q.append('?')
        q.append(query)
    if fragment:
        q.append('#')
        q.append(fragment)
    return (netloc, path + ''.join(q))


def get_or_post(value):
    value = value.upper()
    if value not in ('GET', 'POST'):
        raise ValueError('method must be "GET" or "POST", instead received: '
                         + repr(value))
    return value


class HTTPHandlerFactory(HandlerFactory):

    def create_loghandler(self):
        from ZConfig.components.logger import loghandler
        host, selector = self.section.url
        return loghandler.HTTPHandler(host, selector, self.section.method)


class SMTPHandlerFactory(HandlerFactory):

    def __init__(self, section):
        HandlerFactory.__init__(self, section)
        username = self.section.smtp_username
        password = self.section.smtp_password
        if (username or password) and not (username and password):
            raise ValueError(
                'Either both smtp-username and smtp-password or none must be '
                'given')

    def create_loghandler(self):
        from ZConfig.components.logger import loghandler
        host, port = self.section.smtp_server
        if not port:
            mailhost = host
        else:
            mailhost = host, port
        kwargs = {}
        if self.section.smtp_username:
            kwargs['credentials'] = (self.section.smtp_username,
                                     self.section.smtp_password)
        return loghandler.SMTPHandler(mailhost,
                                      self.section.fromaddr,
                                      self.section.toaddrs,
                                      self.section.subject,
                                      **kwargs)



##############################################################################
#
# Copyright (c) 2018 Zope Foundation and Contributors.
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################
"""\
ZConfig log message formatting support.

"""

import inspect
import logging
import string
import sys


class PercentStyle:

    logging_style = '%'
    default_format = '%(message)s'
    asctime_format = '%(asctime)s'
    asctime_search = '%(asctime)'

    def __init__(self, fmt):
        self._fmt = fmt or self.default_format

    def usesTime(self):
        return self._fmt.find(self.asctime_search) >= 0

    def format(self, record):
        return self._fmt % record.__dict__


class StrFormatStyle(PercentStyle):

    logging_style = '{'
    default_format = '{message}'
    asctime_format = '{asctime}'
    asctime_search = '{asctime'

    __formatter = string.Formatter()

    def format(self, record):
        return self.__formatter.vformat(self._fmt, (), record.__dict__)


class StringTemplateStyle(PercentStyle):

    logging_style = '$'
    default_format = '${message}'
    asctime_format = '${asctime}'
    asctime_search = '${asctime}'

    def __init__(self, fmt):
        self._fmt = fmt or self.default_format
        self._tpl = string.Template(self._fmt)

    def usesTime(self):
        fmt = self._fmt
        return fmt.find('$asctime') >= 0 or fmt.find(self.asctime_format) >= 0

    def format(self, record):
        return self._tpl.substitute(record.__dict__)


class SafeStringTemplateStyle(StringTemplateStyle):

    logging_style = None

    def format(self, record):
        return self._tpl.safe_substitute(record.__dict__)


_control_char_rewrites = {r'\n': '\n', r'\t': '\t', r'\b': '\b',
                          r'\f': '\f', r'\r': '\r'}.items()

_log_format_styles = {
    'classic': PercentStyle,
    'format': StrFormatStyle,
    'template': StringTemplateStyle,
    'safe-template': SafeStringTemplateStyle,
}

_log_format_variables = {
    'name': __name__,
    'levelno': 3,
    'levelname': 'DEBUG',
    'pathname': 'apath',
    'filename': 'afile',
    'module': 'amodule',
    'lineno': 1,
    'created': 1.1,
    'asctime': 'atime',
    'msecs': 1.1,
    'relativeCreated': 1.1,
    'thread': 1,
    'message': 'amessage',
    'process': 1,
    'funcName': 'fname',
    }


def ctrl_char_insert(value):
    for pattern, replacement in _control_char_rewrites:
        value = value.replace(pattern, replacement)
    return value


def escaped_string(value):
    return ctrl_char_insert(value)


def log_format_style(value):
    if value.lower() in _log_format_styles:
        return value.lower()
    raise ValueError(
        'log_format_style must be one of %s; found %r'
        % ((', '.join(repr(v) for v in sorted(_log_format_styles))),
           value)
    )


def resolve(name):
    """Given a dotted name, returns an object imported from a Python module."""
    name = name.split('.')
    used = name.pop(0)
    found = __import__(used)
    for n in name:
        used += '.' + n
        try:
            found = getattr(found, n)
        except AttributeError:
            __import__(used)
            found = getattr(found, n)
    return found


class AnyFieldDict(dict):
    # Only used for format string validation.

    def __getitem__(self, key):
        if key in self:
            # Avoid magic for known keys; better to provide a sample
            # value of an appropriate type when there is such a thing.
            return dict.__getitem__(self, key)
        else:
            # Using an int as the default works with "reasonable" values
            # that might be formatted in a logging configuration:
            #
            # - str/repr are safe
            # - numeric formatting is safe, allowing for casting to float
            #
            return 42


class FormatterFactory:

    def __init__(self, section):
        self.format = section.format
        self.dateformat = section.dateformat
        self.style = section.style
        self.stylist = _log_format_styles[self.style](self.format)
        self.formatter = section.formatter or 'logging.Formatter'
        if section.formatter:
            self.factory = resolve(section.formatter)
        else:
            self.factory = logging.Formatter

        if inspect.isclass(self.factory):
            func = self.factory.__init__
        else:
            func = self.factory
        params = inspect.signature(func).parameters
        self._has_style_param = 'style' in params

        # Check that the format specified complies with the style; we
        # just want the format call to not fall over.  If it does, the
        # exception will propagate and generate a (hopefully)
        # informative error.
        #
        record = logging.LogRecord(__name__, logging.INFO, __file__,
                                   42, 'some message', (), None)
        record.__dict__.update(_log_format_variables)
        if section.arbitrary_fields:
            fields = AnyFieldDict()
            fields.update(record.__dict__)
            record.__dict__ = fields
        try:
            self.stylist.format(record)
        except IndexError:
            #
            # We're checking this exception to catch & report
            # format-style templates that used {} or {#} placeholders,
            # since those aren't allowed when formatting with a mapping.
            #
            raise ValueError('%s formats cannot use positional placeholders')

    def __call__(self):
        #
        # Need to determine if we should pass
        # style=self.message_formatter.logging_style, or if we need to
        # clobber usesTime and formatMessage directly.
        #
        stylist = self.stylist
        if self._has_style_param:
            if stylist.logging_style:
                formatter = self.factory(self.format, self.dateformat,
                                         style=stylist.logging_style)
            else:
                # A formatter class that supports style, but our style is
                # non-standard, so we reach under the covers a bit.
                #
                # Python 3.8 adds a validate option, defaulting to True,
                # which causes the format string to be checked.  Since
                # safe-template is not a standard style, we want to
                # suppress this.
                #
                kwargs = dict()
                if sys.version_info >= (3, 8):
                    kwargs['validate'] = False
                formatter = self.factory(self.format, self.dateformat,
                                         style='$', **kwargs)
                assert formatter._style._fmt == self.format
                formatter._style = stylist
        else:
            formatter = self.factory(self.format, self.dateformat)
            if self.style != 'classic':
                #
                # Should verify that these are defined by
                # logging.Formatter; if not, we risk losing
                # functionality, and should scream so as to avoid
                # surprises.
                #
                formatter.usesTime = stylist.usesTime
                formatter.formatMessage = stylist.format
        return formatter



##############################################################################
#
# Copyright (c) 2019 Zope Corporation and Contributors.
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################
"""\
pygments-compatible lexer.

"""


import pygments.lexer
import pygments.token

import ZConfig.substitution


class ZConfigLexer(pygments.lexer.RegexLexer):
    """pygments lexer for ZConfig configuration syntax."""

    name = 'ZConfig'
    aliases = ['zconfig']

    tokens = {
        'root': [
            (r'\s+', pygments.token.Text),
            (r'(%%define)(\s+)(%s)(\s+)?'
             % ZConfig.substitution._name_re,
             pygments.lexer.bygroups(
                 pygments.token.Keyword, pygments.token.Whitespace,
                 pygments.token.Name, pygments.token.Whitespace),
             'value'),
            (r'(%import)(\s+)',
             pygments.lexer.bygroups(
                 pygments.token.Keyword, pygments.token.Whitespace),
             'value'),
            (r'(%include)(\s+)',
             pygments.lexer.bygroups(
                 pygments.token.Keyword, pygments.token.Whitespace),
             'value'),
            (r'(#.*?)$', pygments.token.Comment),
            (r'(<[^\s>]+)(\s*)(/?>)',
             pygments.lexer.bygroups(
                 pygments.token.Name.Tag, pygments.token.Whitespace,
                 pygments.token.Name.Tag)),
            (r'(<[^\s>]+)(?:(\s+)([^/>]*))?(\s*)(/?>)',
             pygments.lexer.bygroups(
                 pygments.token.Name.Tag, pygments.token.Whitespace,
                 pygments.token.Name, pygments.token.Whitespace,
                 pygments.token.Name.Tag)),
            (r'([a-z]\w*)(\s+)',
             pygments.lexer.bygroups(
                 pygments.token.Name, pygments.token.Whitespace),
             'value'),
            (r'[^\n]+', pygments.token.Text),
        ],
        'value': [
            (r'\$\$', pygments.token.String.Escape),
            (r'(\${)(%s)(})' % ZConfig.substitution._name_re,
             pygments.lexer.bygroups(
                 pygments.token.String.Interpol, pygments.token.Name,
                 pygments.token.String.Interpol)),
            (r'(\$)(%s)' % ZConfig.substitution._name_re,
             pygments.lexer.bygroups(
                 pygments.token.String.Interpol, pygments.token.Name)),
            (r'\n', pygments.token.Text, '#pop'),
            (r'.[^\n$]*', pygments.token.String),
        ],
    }



##############################################################################
#
# Copyright (c) 2002, 2003, 2018 Zope Foundation and Contributors.
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################
"""Structured, schema-driven configuration library.

ZConfig is a configuration library intended for general use.  It
supports a hierarchical schema-driven configuration model that allows
a schema to specify data conversion routines written in Python.
ZConfig's model is very different from the model supported by the
ConfigParser module found in Python's standard library, and is more
suitable to configuration-intensive applications.

ZConfig schema are written in an XML-based language and are able to
``import`` schema components provided by Python packages.  Since
components are able to bind to conversion functions provided by Python
code in the package (or elsewhere), configuration objects can be
arbitrarily complex, with values that have been verified against
arbitrary constraints.  This makes it easy for applications to
separate configuration support from configuration loading even with
configuration data being defined and consumed by a wide range of
separate packages.

"""
__docformat__ = "reStructuredText"

from io import StringIO

import ZConfig.loader


loadConfigFile = ZConfig.loader.loadConfigFile
loadSchemaFile = ZConfig.loader.loadSchemaFile
loadConfig = ZConfig.loader.loadConfig
loadSchema = ZConfig.loader.loadSchema


version_info = (3, 0)
__version__ = ".".join([str(n) for n in version_info])


class ConfigurationError(Exception):
    """Base class for exceptions specific to the :mod:`ZConfig` package.

    All instances provide a ``message`` attribute that describes
    the specific error, and a ``url`` attribute that gives the URL
    of the resource the error was located in, or ``None``.

    """

    # The 'message' attribute was deprecated for BaseException with
    # Python 2.6; here we create descriptor properties to continue using it
    def __set_message(self, v):
        self.__dict__['message'] = v

    def __get_message(self):
        return self.__dict__['message']

    def __del_message(self):
        del self.__dict__['message']

    message = property(__get_message, __set_message, __del_message)

    def __init__(self, msg, url=None):
        self.message = msg
        self.url = url
        Exception.__init__(self, msg)

    def __str__(self):
        return self.message


class _ParseError(ConfigurationError):
    def __init__(self, msg, url, lineno, colno=None):
        self.lineno = lineno
        self.colno = colno
        ConfigurationError.__init__(self, msg, url)

    def __str__(self):
        s = self.message
        if self.url:
            s += "\n("
        elif (self.lineno, self.colno) != (None, None):
            s += " ("
        if self.lineno:
            s += "line %d" % self.lineno
            if self.colno is not None:
                s += ", column %d" % self.colno
            if self.url:
                s += " in %s)" % self.url
            else:
                s += ")"
        elif self.url:
            s += self.url + ")"
        return s


class SchemaError(_ParseError):
    """Raised when a schema contains an error.

    This exception type provides the attributes ``url``, ``lineno``,
    and ``colno``, which provide the source URL, the line number, and
    the column number at which the error was detected. These attributes
    may be ``None`` in some cases.
    """

    def __init__(self, msg, url=None, lineno=None, colno=None):
        _ParseError.__init__(self, msg, url, lineno, colno)


class UnknownDocumentTypeError(SchemaError):
    """
    Raised when the root element of the document being parsed is
    unexpected.
    """


class SchemaResourceError(SchemaError):
    """Raised when there's an error locating a resource required by the
    schema.

    Instances of this exception class add the attributes ``filename``,
    ``package``, and ``path``, which hold the filename searched for
    within the package being loaded, the name of the package, and the
    ``__path__`` attribute of the package itself (or ``None`` if it
    isn't a package or could not be imported).
    """

    def __init__(self, msg, url=None, lineno=None, colno=None,
                 path=None, package=None, filename=None):
        self.filename = filename
        self.package = package
        if path is not None:
            path = path[:]
        self.path = path
        SchemaError.__init__(self, msg, url, lineno, colno)

    def __str__(self):
        s = SchemaError.__str__(self)
        if self.package is not None:
            s += "\n  Package name: " + repr(self.package)
        if self.filename is not None:
            s += "\n  File name: " + repr(self.filename)
        if self.package is not None:
            s += "\n  Package path: " + repr(self.path)
        return s


class ConfigurationSyntaxError(_ParseError):
    """Exception raised when a configuration source does not conform to
    the allowed syntax.

    In addition to the ``message`` and ``url`` attributes, exceptions
    of this type offer the ``lineno`` attribute, which provides the
    line number at which the error was detected.
    """


class DataConversionError(ConfigurationError, ValueError):
    """Raised when a data type conversion fails with :exc:`ValueError`.

    This exception is a subclass of both :exc:`ConfigurationError` and
    :exc:`ValueError`. The :func:`str` of the exception provides the
    explanation from the original :exc:`ValueError`, and the line
    number and URL of the value which provoked the error. The
    following additional attributes are provided:

    ``colno``
       column number at which the value starts, or ``None``
    ``exception``
       the original :exc:`ValueError` instance
    ``lineno``
       line number on which the value starts
    ``message``
      :func:`str` returned by the original :exc:`ValueError`
    ``value``
        original value passed to the conversion function
    ``url``
        URL of the resource providing the value text
    """

    def __init__(self, exception, value, position):
        ConfigurationError.__init__(self, str(exception))
        self.exception = exception
        self.value = value
        self.lineno, self.colno, self.url = position

    def __str__(self):
        s = "{} (line {}".format(self.message, self.lineno)
        if self.colno is not None:
            s += ", %s" % self.colno
        if self.url:
            s += ", in %s)" % self.url
        else:
            s += ")"
        return s


class SubstitutionSyntaxError(ConfigurationError):
    """Raised when interpolation source text contains syntactical errors."""


class SubstitutionReplacementError(ConfigurationSyntaxError, LookupError):
    """Raised when the source text contains references to names which are
    not defined in *mapping*.

    The attributes ``source`` and ``name`` provide the complete source
    text and the name (converted to lower case) for which no replacement
    is defined.
    """

    def __init__(self, source, name, url=None, lineno=None):
        self.source = source
        self.name = name
        ConfigurationSyntaxError.__init__(
            self, "no replacement for " + repr(name), url, lineno)


def configureLoggers(text):
    """Configure one or more loggers from configuration text."""

    schema = ZConfig.loader.loadSchemaFile(StringIO("""
    <schema>
    <import package='ZConfig.components.logger'/>
    <multisection type='logger' name='*' attribute='loggers'/>
    </schema>
    """))

    config, _ = ZConfig.loader.loadConfigFile(schema, StringIO(text))
    for factory in config.loggers:
        factory()



##############################################################################
#
# Copyright (c) 2002, 2003, 2018 Zope Foundation and Contributors.
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################
"""Tests of the configuration data structures and loader."""
import os
import tempfile
import unittest
from io import StringIO

import ZConfig
from ZConfig.tests.support import CONFIG_BASE
from ZConfig.tests.support import TestHelper


class ConfigurationTestCase(TestHelper, unittest.TestCase):

    schema = None

    def get_schema(self):
        if self.schema is None:
            ConfigurationTestCase.schema = ZConfig.loadSchema(
                CONFIG_BASE + "simple.xml")
        return self.schema

    def load(self, relurl, context=None):
        url = CONFIG_BASE + relurl
        self.conf, self.handlers = ZConfig.loadConfig(self.get_schema(), url)
        conf = self.conf
        self.assertTrue(conf.getSectionName() is None)
        self.assertTrue(conf.getSectionType() is None)
        return conf

    def loadtext(self, text):
        sio = StringIO(text)
        return self.loadfile(sio)

    def loadfile(self, file_or_path):
        schema = self.get_schema()
        self.conf, self.handlers = ZConfig.loadConfigFile(schema, file_or_path)
        return self.conf

    def check_simple_gets(self, conf):
        self.assertEqual(conf.empty, '')
        self.assertEqual(conf.int_var, 12)
        self.assertEqual(conf.neg_int, -2)
        self.assertEqual(conf.float_var, 12.02)
        self.assertEqual(conf.var1, 'abc')
        self.assertTrue(conf.true_var_1)
        self.assertTrue(conf.true_var_2)
        self.assertTrue(conf.true_var_3)
        self.assertTrue(not conf.false_var_1)
        self.assertTrue(not conf.false_var_2)
        self.assertTrue(not conf.false_var_3)
        self.assertEqual(conf.list_1, [])
        self.assertEqual(conf.list_2, ['abc'])
        self.assertEqual(conf.list_3, ['abc', 'def', 'ghi'])
        self.assertEqual(conf.list_4, ['[', 'what', 'now?', ']'])

    def test_simple_gets(self):
        conf = self.load("simple.conf")
        self.check_simple_gets(conf)

    def test_type_errors(self):
        Error = ZConfig.DataConversionError
        raises = self.assertRaises
        raises(Error, self.loadtext, "int-var true")
        raises(Error, self.loadtext, "float-var true")
        raises(Error, self.loadtext, "neg-int false")
        raises(Error, self.loadtext, "true-var-1 0")
        raises(Error, self.loadtext, "true-var-1 1")
        with raises(Error) as e:
            self.loadtext("true-var-1 -1")

        # str doesn't fail
        exc = e.exception
        str(exc)
        self.assertIsNone(exc.colno)
        self.assertIsNone(exc.url)

        exc.colno = 1
        exc.url = 'url'
        self.assertIn('url', str(exc))

    def test_simple_sections(self):
        self.schema = ZConfig.loadSchema(CONFIG_BASE + "simplesections.xml")
        conf = self.load("simplesections.conf")
        self.assertEqual(conf.var, "foo")
        # check each interleaved position between sections
        for c in "0123456":
            self.assertEqual(getattr(conf, "var_" + c), "foo-" + c)
        sect = list(sect for sect in conf.sections
                    if sect.getSectionName() == "name")[0]
        self.assertEqual(sect.var, "bar")
        self.assertEqual(sect.var_one, "splat")
        self.assertTrue(sect.var_three is None)
        sect = list(sect for sect in conf.sections
                    if sect.getSectionName() == "delegate")[0]
        self.assertEqual(sect.var, "spam")
        self.assertEqual(sect.var_two, "stuff")
        self.assertTrue(sect.var_three is None)

    def test_include(self):
        conf = self.load("include.conf")
        self.assertEqual(conf.var1, "abc")
        self.assertEqual(conf.var2, "value2")
        self.assertEqual(conf.var3, "value3")
        self.assertEqual(conf.var4, "value")

    def test_includes_with_defines(self):
        self.schema = ZConfig.loadSchemaFile(StringIO("""\
            <schema>
              <key name='refinner' />
              <key name='refouter' />
            </schema>
            """))
        conf = self.load("outer.conf")
        self.assertEqual(conf.refinner, "inner")
        self.assertEqual(conf.refouter, "outer")

    def test_define(self):
        conf = self.load("simple.conf")
        self.assertEqual(conf.getname, "value")
        self.assertEqual(conf.getnametwice, "valuevalue")
        self.assertEqual(conf.getdollars, "$$")
        self.assertEqual(conf.getempty, "xy")
        self.assertEqual(conf.getwords, "abc two words def")

    def test_define_errors(self):
        # doesn't raise if value is equal
        self.loadtext("%define a value\n%define a value\n")

        self.assertRaises(ZConfig.ConfigurationSyntaxError,
                          self.loadtext, "%define\n")
        self.assertRaises(ZConfig.ConfigurationSyntaxError,
                          self.loadtext, "%define abc-def\n")

        self.assertRaises(ZConfig.SubstitutionReplacementError,
                          self.loadtext,
                          "foo $name")

        with self.assertRaises(ZConfig.ConfigurationSyntaxError) as e:
            self.loadtext("%define a value\n%define a other\n")

        # str doesn't throw unexpected exceptions
        exc = e.exception
        self.assertIn('line', str(exc))
        self.assertNotIn('column', str(exc))
        # doesn't have these properties
        self.assertIsNone(exc.colno)
        self.assertIsNone(exc.url)

        # If we fill them in, we get different str output
        exc.colno = 10
        exc.url = 'a url'
        self.assertIn('column', str(exc))

        # There's also a case if we don't have a line number
        exc.lineno = None
        self.assertNotIn('line', str(exc))

    def test_bad_directive(self):
        self.assertRaisesRegex(ZConfig.ConfigurationSyntaxError,
                               'unknown directive',
                               self.loadtext, '%not a directive')

        self.assertRaisesRegex(ZConfig.ConfigurationSyntaxError,
                               'missing or unrecognized',
                               self.loadtext, '%')

    def test_bad_key(self):
        self.assertRaisesRegex(ZConfig.ConfigurationSyntaxError,
                               'malformed configuration data',
                               self.loadtext, '(int-var')

    def test_bad_section(self):
        self.schema = ZConfig.loadSchema(CONFIG_BASE + "simplesections.xml")
        self.assertRaisesRegex(ZConfig.ConfigurationSyntaxError,
                               'unexpected section end',
                               self.loadtext, '</close>')

        self.assertRaisesRegex(ZConfig.ConfigurationSyntaxError,
                               'unbalanced section end',
                               self.loadtext, '<section>\n</close>')

        self.assertRaisesRegex(ZConfig.ConfigurationSyntaxError,
                               'unclosed sections not allowed',
                               self.loadtext, '<section>\n')

        self.assertRaisesRegex(ZConfig.ConfigurationSyntaxError,
                               'malformed section header',
                               self.loadtext, '<section()>\n</close>')

        self.assertRaisesRegex(ZConfig.ConfigurationSyntaxError,
                               'malformed section end',
                               self.loadtext, '<section>\n</section')

        self.assertRaisesRegex(ZConfig.ConfigurationSyntaxError,
                               'malformed section start',
                               self.loadtext, '<section')

        # ConfigLoader.endSection raises this and it is recaught and
        # changed to a SyntaxError
        self.assertRaisesRegex(ZConfig.ConfigurationSyntaxError,
                               "no values for",
                               self.loadtext,
                               "<hasmin foo>\n</hasmin>")

    def test_configuration_error_str(self):

        e = ZConfig.ConfigurationError('message')
        self.assertEqual(e.message, 'message')
        self.assertEqual('message', str(e))

        # We can delete the message, for some reason
        del e.message

    def test_fragment_ident_disallowed(self):
        self.assertRaises(ZConfig.ConfigurationError,
                          self.load, "simplesections.conf#another")

    def test_load_from_fileobj(self):
        sio = StringIO("%define name value\n"
                       "getname x $name y \n")
        cf = self.loadfile(sio)
        self.assertEqual(cf.getname, "x value y")

    def test_load_from_abspath(self):
        fn = self.write_tempfile()
        try:
            self.check_load_from_path(fn)
        finally:
            os.unlink(fn)

    def test_load_from_relpath(self):
        fn = self.write_tempfile()
        dirname, name = os.path.split(fn)
        pwd = os.getcwd()
        try:
            os.chdir(dirname)
            self.check_load_from_path(name)
        finally:
            os.chdir(pwd)
            os.unlink(fn)

    def write_tempfile(self):
        fn = tempfile.mktemp()
        fp = open(fn, "w")
        fp.write("var1 value\n")
        fp.close()
        return fn

    def check_load_from_path(self, path):
        schema = self.get_schema()
        ZConfig.loadConfig(schema, path)


def test_suite():
    return unittest.defaultTestLoader.loadTestsFromName(__name__)


if __name__ == '__main__':
    unittest.main(defaultTest='test_suite')



##############################################################################
#
# Copyright (c) 2002, 2018 Zope Foundation and Contributors.
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################

"""Tests for logging configuration via ZConfig."""

import doctest
import logging
import os
import sys
import unittest
from io import StringIO
from sys import maxsize

import ZConfig
import ZConfig.components.logger.tests.support
import ZConfig.tests.support
from ZConfig.components.logger import datatypes
from ZConfig.components.logger import handlers
from ZConfig.components.logger import loghandler


class CustomFormatter(logging.Formatter):
    def formatException(self, ei):
        """Format and return the exception information as a string.

        This adds helpful advice to the end of the traceback.
        """
        import traceback
        sio = StringIO()
        traceback.print_exception(ei[0], ei[1], ei[2], file=sio)
        return sio.getvalue() + "... Don't panic!"


def read_file(filename):
    with open(filename) as f:
        return f.read()


class TestConfig(ZConfig.components.logger.tests.support.LoggingTestHelper,
                 unittest.TestCase):

    _schematext = """
      <schema>
        <import package='ZConfig.components.logger'/>
        <section type='eventlog' name='*' attribute='eventlog'/>
      </schema>
    """

    def test_config_without_logger(self):
        conf = self.get_config("")
        self.assertTrue(conf.eventlog is None)

    def test_config_without_handlers(self):
        logger = self.check_simple_logger("<eventlog/>")
        # Make sure there's a NullHandler, since a warning gets
        # printed if there are no handlers:
        self.assertEqual(len(logger.handlers), 1)
        self.assertTrue(isinstance(logger.handlers[0], loghandler.NullHandler))

        # And it does nothing
        logger.handlers[0].emit(None)
        logger.handlers[0].handle(None)

    def test_factory_without_stream(self):
        factory = self.check_simple_logger_factory("<eventlog>\n"
                                                   "  <logfile>\n"
                                                   "    path STDERR\n"
                                                   "  </logfile>\n"
                                                   "  <logfile>\n"
                                                   "    path STDERR\n"
                                                   "    level info\n"
                                                   "  </logfile>\n"
                                                   "  <logfile>\n"
                                                   "    path STDERR\n"
                                                   "    level debug\n"
                                                   "  </logfile>\n"
                                                   "</eventlog>")

        factory.startup()
        logger = factory.instance

        factory.level = logging.NOTSET
        self.assertEqual(factory.getLowestHandlerLevel(), logging.DEBUG)
        logger.handlers[0].reopen = lambda: None
        factory.reopen()

    def test_with_logfile(self):
        fn = self.mktemp()
        logger = self.check_simple_logger("<eventlog>\n"
                                          "  <logfile>\n"
                                          "    path %s\n"
                                          "    level debug\n"
                                          "  </logfile>\n"
                                          "</eventlog>" % fn)
        logfile = logger.handlers[0]
        self.assertEqual(logfile.level, logging.DEBUG)
        self.assertTrue(isinstance(logfile, loghandler.FileHandler))
        self.assertFalse(logfile.delay)
        self.assertIsNotNone(logfile.stream)
        logger.removeHandler(logfile)
        logfile.close()

    def test_with_encoded(self):
        fn = self.mktemp()
        logger = self.check_simple_logger("<eventlog>\n"
                                          "  <logfile>\n"
                                          "    path %s\n"
                                          "    level debug\n"
                                          "    encoding shift-jis\n"
                                          "  </logfile>\n"
                                          "</eventlog>" % fn)
        logfile = logger.handlers[0]
        self.assertEqual(logfile.level, logging.DEBUG)
        self.assertTrue(isinstance(logfile, loghandler.FileHandler))
        self.assertFalse(logfile.delay)
        self.assertIsNotNone(logfile.stream)
        self.assertEqual(logfile.stream.encoding, "shift-jis")
        logger.removeHandler(logfile)
        logfile.close()

    def test_with_logfile_delayed(self):
        fn = self.mktemp()
        logger = self.check_simple_logger("<eventlog>\n"
                                          "  <logfile>\n"
                                          "    path %s\n"
                                          "    level debug\n"
                                          "    delay true\n"
                                          "  </logfile>\n"
                                          "</eventlog>" % fn)
        logfile = logger.handlers[0]
        self.assertEqual(logfile.level, logging.DEBUG)
        self.assertTrue(isinstance(logfile, loghandler.FileHandler))
        self.assertTrue(logfile.delay)
        self.assertIsNone(logfile.stream)
        logger.info("this is a test")
        self.assertIsNotNone(logfile.stream)
        logger.removeHandler(logfile)
        logfile.close()

    def test_with_stderr(self):
        self.check_standard_stream("stderr")

    def test_with_stdout(self):
        self.check_standard_stream("stdout")

    def test_delayed_stderr(self):
        self.check_standard_stream_cannot_delay("stderr")

    def test_delayed_stdout(self):
        self.check_standard_stream_cannot_delay("stdout")

    def test_encoded_stderr(self):
        self.check_standard_stream_cannot_encode("stderr")

    def test_encoded_stdout(self):
        self.check_standard_stream_cannot_encode("stdout")

    def test_with_rotating_logfile(self):
        fn = self.mktemp()
        logger = self.check_simple_logger("<eventlog>\n"
                                          "  <logfile>\n"
                                          "    path %s\n"
                                          "    level debug\n"
                                          "    max-size 5mb\n"
                                          "    old-files 10\n"
                                          "  </logfile>\n"
                                          "</eventlog>" % fn)
        logfile = logger.handlers[0]
        self.assertEqual(logfile.level, logging.DEBUG)
        self.assertEqual(logfile.backupCount, 10)
        self.assertEqual(logfile.maxBytes, 5*1024*1024)
        self.assertTrue(isinstance(logfile, loghandler.RotatingFileHandler))
        logger.removeHandler(logfile)
        logfile.close()

    def test_with_timed_rotating_logfile(self):
        fn = self.mktemp()
        logger = self.check_simple_logger("<eventlog>\n"
                                          "  <logfile>\n"
                                          "    path %s\n"
                                          "    level debug\n"
                                          "    when D\n"
                                          "    interval 3\n"
                                          "    old-files 11\n"
                                          "  </logfile>\n"
                                          "</eventlog>" % fn)
        logfile = logger.handlers[0]
        self.assertEqual(logfile.level, logging.DEBUG)
        self.assertEqual(logfile.backupCount, 11)
        self.assertEqual(logfile.interval, 86400*3)
        self.assertIsInstance(logfile, loghandler.TimedRotatingFileHandler)
        logger.removeHandler(logfile)
        logfile.close()

    def test_with_timed_rotating_logfile_and_size_should_fail(self):
        fn = self.mktemp()
        self.assertRaises(
            ZConfig.DataConversionError,
            self.check_simple_logger_factory,
            "<eventlog>\n"
            "  <logfile>\n"
            "    path %s\n"
            "    level debug\n"
            "    max-size 5mb\n"
            "    when D\n"
            "    old-files 10\n"
            "  </logfile>\n"
            "</eventlog>" % fn)

        # Mising old-files
        self.assertRaisesRegex(
            ZConfig.DataConversionError,
            "old-files must be set",
            self.check_simple_logger_factory,
            "<eventlog>\n"
            "  <logfile>\n"
            "    path %s\n"
            "    level debug\n"
            "    max-size 5mb\n"
            "    when D\n"
            "  </logfile>\n"
            "</eventlog>" % fn)

        self.assertRaisesRegex(
            ZConfig.DataConversionError,
            "max-bytes or when must be set",
            self.check_simple_logger_factory,
            "<eventlog>\n"
            "  <logfile>\n"
            "    path %s\n"
            "    level debug\n"
            "    interval 1\n"
            "    old-files 10\n"
            "  </logfile>\n"
            "</eventlog>" % fn)

    def test_with_rotating_logfile_and_STD_should_fail(self):
        for path in ('STDERR', 'STDOUT'):
            for param in ('old-files 10', 'max-size 5mb'):
                self.assertRaises(
                    ZConfig.DataConversionError,
                    self.check_simple_logger_factory,
                    "<eventlog>\n"
                    "  <logfile>\n"
                    "    path %s\n"
                    "    level debug\n"
                    "    when D\n"
                    "    %s\n"
                    "  </logfile>\n"
                    "</eventlog>" % (path, param))

    def check_standard_stream(self, name):
        old_stream = getattr(sys, name)
        conf = self.get_config("""
            <eventlog>
              <logfile>
                level info
                path %s
              </logfile>
            </eventlog>
            """ % name.upper())
        self.assertTrue(conf.eventlog is not None)
        # The factory has already been created; make sure it picks up
        # the stderr we set here when we create the logger and
        # handlers:
        sio = StringIO()
        setattr(sys, name, sio)
        try:
            logger = conf.eventlog()
        finally:
            setattr(sys, name, old_stream)
        logger.warning("woohoo!")
        self.assertIs(logger.handlers[0].stream, sio)
        self.assertTrue(sio.getvalue().find("woohoo!") >= 0)

    def check_standard_stream_cannot_delay(self, name):
        with self.assertRaises(ZConfig.DataConversionError) as cm:
            self.get_config("""
                <eventlog>
                  <logfile>
                    level info
                    path %s
                    delay true
                  </logfile>
                </eventlog>
                """ % name.upper())

        self.assertIn("cannot delay opening %s" % name.upper(),
                      str(cm.exception))

    def check_standard_stream_cannot_encode(self, name):
        with self.assertRaises(ZConfig.DataConversionError) as cm:
            self.get_config("""
                <eventlog>
                  <logfile>
                    level info
                    path %s
                    encoding utf-8
                  </logfile>
                </eventlog>
                """ % name.upper())
        self.assertIn("cannot specify encoding for %s" % name.upper(),
                      str(cm.exception))

    def test_custom_formatter(self):
        clsname = __name__ + '.CustomFormatter'
        old_stream = sys.stdout
        sio = StringIO()
        sys.stdout = sio
        try:
            conf = self.get_config("""
                <eventlog>
                  <logfile>
                    formatter %s
                    level info
                    path STDOUT
                  </logfile>
                </eventlog>
                """ % clsname)
            logger = conf.eventlog()
        finally:
            sys.stdout = old_stream
        try:
            raise KeyError
        except KeyError:
            logger.exception("testing a KeyError")
        self.assertTrue(sio.getvalue().find("KeyError") >= 0)
        self.assertTrue(sio.getvalue().find("Don't panic") >= 0)

    def test_with_syslog(self):
        logger = self.check_simple_logger("<eventlog>\n"
                                          "  <syslog>\n"
                                          "    level error\n"
                                          "    facility local3\n"
                                          "  </syslog>\n"
                                          "</eventlog>")
        syslog = logger.handlers[0]
        self.assertEqual(syslog.level, logging.ERROR)
        self.assertTrue(isinstance(syslog, loghandler.SysLogHandler))
        syslog.close()  # avoid ResourceWarning

    def test_with_http_logger_localhost(self):
        logger = self.check_simple_logger("<eventlog>\n"
                                          "  <http-logger>\n"
                                          "    level error\n"
                                          "    method post\n"
                                          "  </http-logger>\n"
                                          "</eventlog>")
        handler = logger.handlers[0]
        self.assertEqual(handler.host, "localhost")
        # XXX The "url" attribute of the handler is misnamed; it
        # really means just the selector portion of the URL.
        self.assertEqual(handler.url, "/")
        self.assertEqual(handler.level, logging.ERROR)
        self.assertEqual(handler.method, "POST")
        self.assertTrue(isinstance(handler, loghandler.HTTPHandler))

    def test_with_http_logger_remote_host(self):
        logger = self.check_simple_logger("<eventlog>\n"
                                          "  <http-logger>\n"
                                          "    method get\n"
                                          "    url http://example.com/log/\n"
                                          "  </http-logger>\n"
                                          "</eventlog>")
        handler = logger.handlers[0]
        self.assertEqual(handler.host, "example.com")
        # XXX The "url" attribute of the handler is misnamed; it
        # really means just the selector portion of the URL.
        self.assertEqual(handler.url, "/log/")
        self.assertEqual(handler.level, logging.NOTSET)
        self.assertEqual(handler.method, "GET")
        self.assertTrue(isinstance(handler, loghandler.HTTPHandler))

    def test_with_email_notifier(self):
        logger = self.check_simple_logger("<eventlog>\n"
                                          "  <email-notifier>\n"
                                          "    to sysadmin@example.com\n"
                                          "    to sa-pager@example.com\n"
                                          "    from zlog-user@example.com\n"
                                          "    level fatal\n"
                                          "  </email-notifier>\n"
                                          "</eventlog>")
        handler = logger.handlers[0]
        self.assertEqual(handler.toaddrs, ["sysadmin@example.com",
                                           "sa-pager@example.com"])
        self.assertEqual(handler.fromaddr, "zlog-user@example.com")
        self.assertEqual(handler.level, logging.FATAL)

    def test_with_email_notifier_with_credentials(self):
        logger = self.check_simple_logger("<eventlog>\n"
                                          "  <email-notifier>\n"
                                          "    to sysadmin@example.com\n"
                                          "    from zlog-user@example.com\n"
                                          "    level fatal\n"
                                          "    smtp-server foo:487\n"
                                          "    smtp-username john\n"
                                          "    smtp-password johnpw\n"
                                          "  </email-notifier>\n"
                                          "</eventlog>")

        handler = logger.handlers[0]
        self.assertEqual(handler.toaddrs, ["sysadmin@example.com"])
        self.assertEqual(handler.fromaddr, "zlog-user@example.com")
        self.assertEqual(handler.fromaddr, "zlog-user@example.com")
        self.assertEqual(handler.level, logging.FATAL)
        self.assertEqual(handler.username, 'john')
        self.assertEqual(handler.password, 'johnpw')
        self.assertEqual(handler.mailhost, 'foo')
        self.assertEqual(handler.mailport, 487)

    def test_with_email_notifier_with_invalid_credentials(self):
        # smtp-username without smtp-password
        with self.assertRaises(ZConfig.DataConversionError) as cm:
            self.check_simple_logger_factory(
                "<eventlog>\n"
                "  <email-notifier>\n"
                "    to sysadmin@example.com\n"
                "    from zlog-user@example.com\n"
                "    level fatal\n"
                "    smtp-username john\n"
                "  </email-notifier>\n"
                "</eventlog>")
        self.assertIn(
            'both smtp-username and smtp-password or none must be given',
            str(cm.exception))

        # smtp-password without smtp-username
        with self.assertRaises(ZConfig.DataConversionError) as cm:
            self.check_simple_logger_factory(
                "<eventlog>\n"
                "  <email-notifier>\n"
                "    to sysadmin@example.com\n"
                "    from zlog-user@example.com\n"
                "    level fatal\n"
                "    smtp-password john\n"
                "  </email-notifier>\n"
                "</eventlog>")
        self.assertIn(
            'both smtp-username and smtp-password or none must be given',
            str(cm.exception))

    def check_simple_logger_factory(self, text, level=logging.INFO):
        conf = self.get_config(text)
        self.assertTrue(conf.eventlog is not None)
        self.assertEqual(conf.eventlog.level, level)
        return conf.eventlog

    def check_simple_logger(self, text, level=logging.INFO):
        logger = self.check_simple_logger_factory(text, level)()
        self.assertTrue(isinstance(logger, logging.Logger))
        self.assertEqual(len(logger.handlers), 1)
        return logger


if os.name == 'nt':
    # Though log files can be closed and re-opened on Windows, these
    # tests expect to be able to move the underlying files out from
    # underneath the logger while open.  That's not possible on
    # Windows. So we don't extend TestCase so that they don't get run.
    #
    # Different tests are needed that only test that close/re-open
    # operations are performed by the handler; those can be run on
    # any platform.
    _RotateTestBase = object
else:
    _RotateTestBase = unittest.TestCase


class TestReopeningRotatingLogfiles(
        ZConfig.components.logger.tests.support.LoggingTestHelper,
        _RotateTestBase):

    # These tests should not be run on Windows.

    handler_factory = loghandler.RotatingFileHandler

    _schematext = """
      <schema>
        <import package='ZConfig.components.logger'/>
        <multisection type='logger' name='*' attribute='loggers'/>
      </schema>
    """

    _sampleconfig_template = """
      <logger>
        name  foo.bar
        <logfile>
          path  %(path0)s
          level debug
          max-size 1mb
          old-files 10
        </logfile>
        <logfile>
          path  %(path1)s
          level info
          max-size 1mb
          old-files 3
        </logfile>
        <logfile>
          path  %(path1)s
          level info
          when D
          old-files 3
        </logfile>
      </logger>

      <logger>
        name  bar.foo
        <logfile>
          path  %(path2)s
          level info
          max-size 10mb
          old-files 10
        </logfile>
      </logger>
    """

    def test_filehandler_reopen(self):

        def mkrecord(msg):
            args = ["foo.bar", logging.ERROR, __file__, 42, msg, (), ()]
            return logging.LogRecord(*args)

        # This goes through the reopening operation *twice* to make
        # sure that we don't lose our handle on the handler the first
        # time around.

        fn = self.mktemp()
        h = self.handler_factory(fn)
        h.handle(mkrecord("message 1"))
        nfn1 = self.move(fn)
        h.handle(mkrecord("message 2"))
        h.reopen()
        h.handle(mkrecord("message 3"))
        nfn2 = self.move(fn)
        h.handle(mkrecord("message 4"))
        h.reopen()
        h.handle(mkrecord("message 5"))
        h.close()

        # Check that the messages are in the right files::
        text1 = read_file(nfn1)
        text2 = read_file(nfn2)
        text3 = read_file(fn)
        self.assertTrue("message 1" in text1)
        self.assertTrue("message 2" in text1)
        self.assertTrue("message 3" in text2)
        self.assertTrue("message 4" in text2)
        self.assertTrue("message 5" in text3)

    def test_logfile_reopening(self):
        #
        # This test only applies to the simple logfile reopening; it
        # doesn't work the same way as the rotating logfile handler.
        #
        paths = self.mktemp(), self.mktemp(), self.mktemp()
        d = {
            "path0": paths[0],
            "path1": paths[1],
            "path2": paths[2],
            }
        text = self._sampleconfig_template % d
        conf = self.get_config(text)
        self.assertEqual(len(conf.loggers), 2)
        # Build the loggers from the configuration, and write to them:
        conf.loggers[0]().info("message 1")
        conf.loggers[1]().info("message 2")
        #
        # We expect this to re-open the original filenames, so we'll
        # have six files instead of three.
        #
        loghandler.reopenFiles()
        #
        # Write to them again:
        conf.loggers[0]().info("message 3")
        conf.loggers[1]().info("message 4")
        #
        # We expect this to re-open the original filenames, so we'll
        # have nine files instead of six.
        #
        loghandler.reopenFiles()
        #
        # Write to them again:
        conf.loggers[0]().info("message 5")
        conf.loggers[1]().info("message 6")
        #
        # We should now have all nine files:
        for fn in paths:
            fn1 = fn + ".1"
            fn2 = fn + ".2"
            self.assertTrue(os.path.isfile(fn), "%r must exist" % fn)
            self.assertTrue(os.path.isfile(fn1), "%r must exist" % fn1)
            self.assertTrue(os.path.isfile(fn2), "%r must exist" % fn2)
        #
        # Clean up:
        for logger in conf.loggers:
            logger = logger()
            for handler in logger.handlers[:]:
                logger.removeHandler(handler)
                handler.close()


class TestReopeningLogfiles(TestReopeningRotatingLogfiles):

    handler_factory = loghandler.FileHandler

    _sampleconfig_template = """
      <logger>
        name  foo.bar
        <logfile>
          path  %(path0)s
          level debug
        </logfile>
        <logfile>
          path  %(path1)s
          level info
        </logfile>
      </logger>

      <logger>
        name  bar.foo
        <logfile>
          path  %(path2)s
          level info
        </logfile>
      </logger>
    """

    def test_logfile_reopening(self):
        #
        # This test only applies to the simple logfile reopening; it
        # doesn't work the same way as the rotating logfile handler.
        #
        paths = self.mktemp(), self.mktemp(), self.mktemp()
        d = {
            "path0": paths[0],
            "path1": paths[1],
            "path2": paths[2],
            }
        text = self._sampleconfig_template % d
        conf = self.get_config(text)
        self.assertEqual(len(conf.loggers), 2)
        # Build the loggers from the configuration, and write to them:
        conf.loggers[0]().info("message 1")
        conf.loggers[1]().info("message 2")
        npaths1 = [self.move(fn) for fn in paths]
        #
        # We expect this to re-open the original filenames, so we'll
        # have six files instead of three.
        #
        loghandler.reopenFiles()
        #
        # Write to them again:
        conf.loggers[0]().info("message 3")
        conf.loggers[1]().info("message 4")
        npaths2 = [self.move(fn) for fn in paths]
        #
        # We expect this to re-open the original filenames, so we'll
        # have nine files instead of six.
        #
        loghandler.reopenFiles()
        #
        # Write to them again:
        conf.loggers[0]().info("message 5")
        conf.loggers[1]().info("message 6")
        #
        # We should now have all nine files:
        for fn in paths:
            self.assertTrue(os.path.isfile(fn), "%r must exist" % fn)
        for fn in npaths1:
            self.assertTrue(os.path.isfile(fn), "%r must exist" % fn)
        for fn in npaths2:
            self.assertTrue(os.path.isfile(fn), "%r must exist" % fn)
        #
        # Clean up:
        for logger in conf.loggers:
            logger = logger()
            for handler in logger.handlers[:]:
                logger.removeHandler(handler)
                handler.close()

    def test_filehandler_reopen_thread_safety(self):
        # The reopen method needs to do locking to avoid a race condition
        # with emit calls. For simplicity we replace the "acquire" and
        # "release" methods with dummies that record calls to them.

        fn = self.mktemp()
        h = self.handler_factory(fn)

        calls = []
        h.acquire = lambda: calls.append("acquire")
        h.release = lambda: calls.append("release")

        h.reopen()
        self.assertEqual(calls, ["acquire", "release"])
        del calls[:]

        # FileHandler.close() does acquire/release, and invokes
        # StreamHandler.flush(), which does the same. Since the lock is
        # recursive, that's ok.
        #
        h.close()
        self.assertEqual(calls, ["acquire", "acquire", "release", "release"])

    def test_with_logfile_delayed_reopened(self):
        fn = self.mktemp()
        conf = self.get_config("<logger>\n"
                               "  <logfile>\n"
                               "    path %s\n"
                               "    level debug\n"
                               "    delay true\n"
                               "    encoding shift-jis\n"
                               "  </logfile>\n"
                               "</logger>" % fn)
        logger = conf.loggers[0]()
        logfile = logger.handlers[0]
        self.assertTrue(logfile.delay)
        self.assertIsNone(logfile.stream)
        logger.info("this is a test")
        self.assertIsNotNone(logfile.stream)
        self.assertEqual(logfile.stream.encoding, "shift-jis")

        # After reopening, we expect the stream to be reset, to be
        # opened at the next event to be logged:
        logfile.reopen()
        self.assertIsNone(logfile.stream)
        logger.info("this is another test")
        self.assertIsNotNone(logfile.stream)
        self.assertEqual(logfile.stream.encoding, "shift-jis")
        logger.removeHandler(logfile)
        logfile.close()


class TestFunctions(ZConfig.tests.support.TestHelper, unittest.TestCase):

    def test_log_format_bad_syntax_1(self):
        with self.assertRaises(ValueError) as cm:
            # Really, disallowed character following '%':
            handlers.log_format('%{no-such-key}s')
        self.assertEqual(str(cm.exception),
                         'Invalid log format string %{no-such-key}s')

    def test_log_format_bad_syntax_2(self):
        with self.assertRaises(ValueError) as cm:
            # Missing trailing 's':
            handlers.log_format('%(levelname)')
        self.assertEqual(str(cm.exception),
                         'Invalid log format string %(levelname)')

    def test_log_format_unknown_key(self):
        with self.assertRaises(ValueError) as cm:
            handlers.log_format('%(no-such-key)s')
        self.assertEqual(str(cm.exception),
                         'Invalid log format string %(no-such-key)s')

    def test_log_format_ok(self):
        fmt = handlers.log_format(r'\n\t\b\f\r')
        self.assertEqual(fmt, '\n\t\b\f\r')

    def test_log_format_func_name(self):
        fmt = '%(funcName)s'
        self.assertEqual(handlers.log_format(fmt), fmt)

    def test_log_format_levelno_integer(self):
        fmt = '%(levelno)2d'
        self.assertEqual(handlers.log_format(fmt), fmt)

    def test_log_format_msecs_float(self):
        fmt = '%(msecs)03.0f'
        self.assertEqual(handlers.log_format(fmt), fmt)

    def test_log_format_relative_created_float(self):
        fmt = '%(relativeCreated)+.3f'
        self.assertEqual(handlers.log_format(fmt), fmt)

    def test_resolve_deep(self):
        old_mod = None
        if hasattr(logging, 'handlers'):
            # This module is nested so it hits our coverage target,
            # and it doesn't alter any state
            # on import, so a "reimport" is fine
            del logging.handlers
            old_mod = sys.modules['logging.handlers']
            del sys.modules['logging.handlers']
        try:
            handlers.resolve('logging.handlers')
        finally:
            if old_mod is not None:
                logging.handlers = old_mod
                sys.modules['logging.handlers'] = old_mod

    def test_http_handler_url(self):
        self.assertRaisesRegex(ValueError,
                               'must be an http',
                               handlers.http_handler_url, 'file://foo/baz')
        self.assertRaisesRegex(ValueError,
                               'must specify a location',
                               handlers.http_handler_url, 'http://')
        self.assertRaisesRegex(ValueError,
                               'must specify a path',
                               handlers.http_handler_url, 'http://server')

        v = handlers.http_handler_url("http://server/path;param?q=v#fragment")
        self.assertEqual(v, ('server', '/path;param?q=v#fragment'))

    def test_close_files(self):
        class F:
            closed = 0

            def close(self):
                self.closed += 1

        f = F()

        def wr():
            return f

        loghandler._reopenable_handlers.append(wr)
        loghandler.closeFiles()
        loghandler.closeFiles()

        self.assertEqual(1, f.closed)

    def test_reopen_files_missing_wref(self):
        # simulate concurrent iteration that pops the ref
        def wr():
            loghandler._reopenable_handlers.remove(wr)

        loghandler._reopenable_handlers.append(wr)
        loghandler.reopenFiles()

    def test_logging_level(self):
        # Make sure the expected names are supported; it's not clear
        # how to check the values in a meaningful way.
        # Just make sure they're case-insensitive.
        convert = datatypes.logging_level
        for name in ["notset", "all", "trace", "debug", "blather",
                     "info", "warn", "warning", "error", "fatal",
                     "critical"]:
            self.assertEqual(convert(name), convert(name.upper()))
        self.assertRaises(ValueError, convert, "hopefully-not-a-valid-value")
        self.assertEqual(convert('10'), 10)
        self.assertRaises(ValueError, convert, '-1000')
        self.assertRaises(ValueError, convert, '-1')
        self.assertRaises(ValueError, convert, '51')
        self.assertRaises(ValueError, convert, '100')

    def test_http_method(self):
        convert = handlers.get_or_post
        self.assertEqual(convert("get"), "GET")
        self.assertEqual(convert("GET"), "GET")
        self.assertEqual(convert("post"), "POST")
        self.assertEqual(convert("POST"), "POST")
        self.assertRaises(ValueError, convert, "")
        self.assertRaises(ValueError, convert, "foo")

    def test_syslog_facility(self):
        convert = handlers.syslog_facility
        for name in ["auth", "authpriv", "cron", "daemon", "kern",
                     "lpr", "mail", "news", "security", "syslog",
                     "user", "uucp", "local0", "local1", "local2",
                     "local3", "local4", "local5", "local6", "local7"]:
            self.assertEqual(convert(name), name)
            self.assertEqual(convert(name.upper()), name)
        self.assertRaises(ValueError, convert, "hopefully-never-a-valid-value")


class TestStartupHandler(unittest.TestCase):

    def test_buffer(self):
        handler = loghandler.StartupHandler()
        self.assertFalse(handler.shouldFlush(None))
        self.assertEqual(maxsize, handler.capacity)

        records = []

        def handle(record):
            records.append(record)

        handle.handle = handle

        handler.flushBufferTo(handle)
        self.assertEqual([], records)

        handler.buffer.append(1)
        handler.flushBufferTo(handle)
        self.assertEqual([1], records)

        del handle.handle


def test_logger_convenience_function_and_ommiting_name_to_get_root_logger():
    """

The ZConfig.loggers function can be used to configure one or more loggers.
We'll configure the rot logger and a non-root logger.

    >>> old_level = logging.getLogger().getEffectiveLevel()
    >>> old_handler_count = len(logging.getLogger().handlers)

    >>> ZConfig.configureLoggers('''
    ... <logger>
    ...    level INFO
    ...    <logfile>
    ...       PATH STDOUT
    ...       format root %(levelname)s %(name)s %(message)s
    ...    </logfile>
    ... </logger>
    ...
    ... <logger>
    ...    name ZConfig.TEST
    ...    level DEBUG
    ...    <logfile>
    ...       PATH STDOUT
    ...       format test %(levelname)s %(name)s %(message)s
    ...    </logfile>
    ... </logger>
    ... ''')

    >>> logging.getLogger('ZConfig.TEST').debug('test message')
    test DEBUG ZConfig.TEST test message
    root DEBUG ZConfig.TEST test message

    >>> logging.getLogger().getEffectiveLevel() == logging.INFO
    True
    >>> len(logging.getLogger().handlers) == old_handler_count + 1
    True
    >>> logging.getLogger('ZConfig.TEST').getEffectiveLevel() == logging.DEBUG
    True
    >>> len(logging.getLogger('ZConfig.TEST').handlers) == 1
    True

.. cleanup

    >>> logging.getLogger('ZConfig.TEST').setLevel(logging.NOTSET)
    >>> logging.getLogger('ZConfig.TEST').removeHandler(
    ...     logging.getLogger('ZConfig.TEST').handlers[-1])
    >>> logging.getLogger().setLevel(old_level)
    >>> logging.getLogger().removeHandler(logging.getLogger().handlers[-1])


    """


def test_suite():
    return unittest.TestSuite([
        unittest.defaultTestLoader.loadTestsFromName(__name__),
        doctest.DocTestSuite()
    ])


if __name__ == '__main__':
    unittest.main(defaultTest="test_suite")



##############################################################################
#
# Copyright (c) 2009 Zope Foundation and Contributors.
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################
import doctest
import logging
import os
import os.path
import unittest

import manuel.capture
import manuel.doctest
import manuel.testing


options = doctest.REPORT_NDIFF | doctest.ELLIPSIS

old = {}


def setUp(test):
    logger = logging.getLogger()
    old['level'] = logger.level
    old['handlers'] = logger.handlers[:]


def tearDown(test):
    logger = logging.getLogger()
    logger.level = old['level']
    logger.handlers = old['handlers']


def findRoot():
    here = os.path.dirname(os.path.abspath(__file__))
    while not os.path.exists(os.path.join(here, 'setup.py')):
        prev, here = here, os.path.dirname(here)
        if here == prev:
            # Let's avoid infinite loops at root
            raise AssertionError('could not find my setup.py')
    return here


def docSetUp(test):
    # Python makes __path__ and __file__ relative in some
    # cases (such as when we're executing with the 'ZConfig'
    # directory on sys.path as CWD). This breaks finding
    # schema components when we change directories.
    import ZConfig.components.logger as logger
    logger.__file__ = os.path.abspath(logger.__file__)
    logger.__path__ = [os.path.abspath(x) for x in logger.__path__]

    old['pwd'] = os.getcwd()
    doc_path = os.path.join(
        findRoot(),
        'docs')
    os.chdir(doc_path)
    setUp(test)


def docTearDown(test):
    os.chdir(old['pwd'])
    tearDown(test)
    old.clear()


def test_suite():
    root = findRoot()
    plugins = manuel.doctest.Manuel(optionflags=options)
    plugins += manuel.capture.Manuel()
    return unittest.TestSuite([
        manuel.testing.TestSuite(
            plugins,
            os.path.join(root, 'README.rst'),
            setUp=setUp, tearDown=tearDown,
        ),
        manuel.testing.TestSuite(
            plugins,
            os.path.join(root, 'docs', 'using-logging.rst'),
            globs={'resetLoggers': lambda: tearDown(None)},
            setUp=docSetUp, tearDown=docTearDown,
        ),
    ])


if __name__ == '__main__':
    unittest.main(defaultTest='test_suite')



##############################################################################
#
# Copyright (c) 2007 Zope Foundation and Contributors.
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################
"""\
Test driver for ZConfig.schemaless.

"""
__docformat__ = "reStructuredText"

import unittest

import manuel.doctest
import manuel.testing

from ZConfig.schemaless import Section


class TestSection(unittest.TestCase):

    def test_init_with_data(self):
        s = Section(data={'k': 'v'})
        self.assertDictEqual(s, {'k': 'v'})


def test_suite():
    return unittest.TestSuite([
        unittest.defaultTestLoader.loadTestsFromName(__name__),
        manuel.testing.TestSuite(
            manuel.doctest.Manuel(),
            '../schemaless.txt'),
    ])


if __name__ == '__main__':
    unittest.main(defaultTest='test_suite')



##############################################################################
#
# Copyright (c) 2002, 2003, 2018 Zope Foundation and Contributors.
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################
"""Default implementation of a data type registry

This module provides the implementation of the default data type
registry and all the standard data types supported by :mod:`ZConfig`.
A number of convenience classes are also provided to assist in the
creation of additional data types.

A "data type registry" is an object that provides conversion functions
for data types. The interface for a :class:`registry <Registry>` is
fairly simple.

A "conversion function" is any callable object that accepts a single
argument and returns a suitable value, or raises an exception if the
input value is not acceptable. :exc:`ValueError` is the preferred
exception for disallowed inputs, but any other exception will be
properly propagated.

"""

import datetime
import os
import re
import sys
from functools import reduce


class MemoizedConversion:
    """Simple memoization for potentially expensive conversions.

    This conversion helper caches each successful conversion for re-use
    at a later time; failed conversions are not cached in any way, since
    it is difficult to raise a meaningful exception providing
    information about the specific failure.

    """

    def __init__(self, conversion):
        self._memo = {}
        self._conversion = conversion

    def __call__(self, value):
        try:
            return self._memo[value]
        except KeyError:
            v = self._conversion(value)
            self._memo[value] = v
            return v


class RangeCheckedConversion:
    """Conversion helper that performs range checks on the result of
    another conversion.

    Values passed to instances of this conversion are converted using
    *conversion* and then range checked. *min* and *max*, if given and
    not ``None``, are the inclusive endpoints of the allowed range.
    Values returned by *conversion* which lay outside the range
    described by *min* and *max* cause :exc:`ValueError` to be raised.

    """

    def __init__(self, conversion, min=None, max=None):
        self._min = min
        self._max = max
        self._conversion = conversion

    def __call__(self, value):
        v = self._conversion(value)
        if self._min is not None and v < self._min:
            raise ValueError("%s is below lower bound (%s)"
                             % (repr(v), repr(self._min)))
        if self._max is not None and v > self._max:
            raise ValueError("%s is above upper bound (%s)"
                             % (repr(v), repr(self._max)))
        return v


class RegularExpressionConversion:
    """Conversion that checks that the input matches the regular
    expression *regex*.

    If it matches, returns the input, otherwise raises
    :exc:`ValueError`.
    """

    reason = "value did not match regular expression"

    def __init__(self, regex):
        self._rx = re.compile(regex)

    def __call__(self, value):
        m = self._rx.match(value)
        if m and m.group() == value:
            return value
        else:
            raise ValueError("{}: {}".format(self.reason, repr(value)))


def check_locale(value):
    import locale
    prev = locale.setlocale(locale.LC_ALL)
    try:
        try:
            locale.setlocale(locale.LC_ALL, value)
        finally:
            locale.setlocale(locale.LC_ALL, prev)
    except locale.Error:
        raise ValueError(
            'The specified locale "%s" is not supported by your system.\n'
            'See your operating system documentation for more\n'
            'information on locale support.' % value)
    else:
        return value


class BasicKeyConversion(RegularExpressionConversion):

    def __init__(self):
        RegularExpressionConversion.__init__(self, "[a-zA-Z][-._a-zA-Z0-9]*")

    def __call__(self, value):
        value = str(value)
        return RegularExpressionConversion.__call__(self, value).lower()


class ASCIIConversion(RegularExpressionConversion):
    pass


_ident_re = "[_a-zA-Z][_a-zA-Z0-9]*"


class IdentifierConversion(ASCIIConversion):
    reason = "not a valid Python identifier"

    def __init__(self):
        ASCIIConversion.__init__(self, _ident_re)


class DottedNameConversion(ASCIIConversion):
    reason = "not a valid dotted name"

    def __init__(self):
        ASCIIConversion.__init__(self,
                                 r"{}(?:\.{})*".format(_ident_re, _ident_re))


class DottedNameSuffixConversion(ASCIIConversion):
    reason = "not a valid dotted name or suffix"

    def __init__(self):
        ASCIIConversion.__init__(self,
                                 r"(?:%s)(?:\.%s)*|(?:\.%s)+"
                                 % (_ident_re, _ident_re, _ident_re))


def integer(value):
    return int(value)


def null_conversion(value):
    return value


def asBoolean(s):
    """Convert a string value to a boolean value."""
    ss = str(s).lower()
    if ss in ('yes', 'true', 'on'):
        return True
    elif ss in ('no', 'false', 'off'):
        return False
    else:
        raise ValueError("not a valid boolean value: " + repr(s))


def string_list(s):
    """Convert a string to a list of strings using .split()."""
    return s.split()


port_number = RangeCheckedConversion(integer, min=0, max=0xffff).__call__


class InetAddress:

    def __init__(self, default_host):
        self.DEFAULT_HOST = default_host

    def __call__(self, s):
        # returns (host, port) tuple
        host = ''
        port = None
        if ":" in s:
            host, p = s.rsplit(":", 1)
            if host.startswith('[') and host.endswith(']'):
                # [IPv6]:port
                host = host[1:-1]
            elif ':' in host:
                # Unbracketed IPv6 address;
                # last part is not the port number
                host = s
                p = None
            if p:  # else leave port at None
                port = port_number(p)
            host = host.lower()
        else:
            try:
                port = port_number(s)
            except ValueError:
                if len(s.split()) != 1:
                    raise ValueError("not a valid host name: " + repr(s))
                host = s.lower()
        if not host:
            host = self.DEFAULT_HOST
        return host, port


if sys.platform[:3] == "win":
    DEFAULT_HOST = "localhost"
else:
    DEFAULT_HOST = ""

inet_address = InetAddress(DEFAULT_HOST)
inet_connection_address = InetAddress("127.0.0.1")
inet_binding_address = InetAddress("")


class SocketAddress:
    # Parsing results in family and address
    # Family can be AF_UNIX (for addresses that are path names)
    # or AF_INET6 (for inet addresses with colons in them)
    # or AF_INET (for all other inet addresses);
    # An inet address is a (host, port) pair
    # Notice that no DNS lookup is performed, so if the host
    # is a DNS name, DNS lookup may end up with either IPv4 or
    # IPv6 addresses, or both

    def __init__(self, s):
        import socket
        if "/" in s or s.find(os.sep) >= 0:
            self.family = getattr(socket, "AF_UNIX", None)
            self.address = s
        else:
            self.family = socket.AF_INET
            self.address = self._parse_address(s)
            if ':' in self.address[0]:
                self.family = socket.AF_INET6

    def _parse_address(self, s):
        return inet_address(s)


class SocketBindingAddress(SocketAddress):

    def _parse_address(self, s):
        return inet_binding_address(s)


class SocketConnectionAddress(SocketAddress):

    def _parse_address(self, s):
        return inet_connection_address(s)


def float_conversion(v):
    return float(v)


class IpaddrOrHostname(RegularExpressionConversion):
    def __init__(self):
        # IP address regex from the Perl Cookbook, Recipe 6.23 (revised ed.)
        # We allow underscores in hostnames although this is considered
        # illegal according to RFC1034.
        # Addition: IPv6 addresses are now also accepted
        expr = (r"(^(\d|[01]?\d\d|2[0-4]\d|25[0-5])\."  # ipaddr
                r"(\d|[01]?\d\d|2[0-4]\d|25[0-5])\."    # ipaddr cont'd
                r"(\d|[01]?\d\d|2[0-4]\d|25[0-5])\."    # ipaddr cont'd
                r"(\d|[01]?\d\d|2[0-4]\d|25[0-5])$)"    # ipaddr cont'd
                r"|([A-Za-z_][-A-Za-z0-9_.]*[-A-Za-z0-9_])"  # or hostname
                # or superset of IPv6 addresses (requiring at least one colon)
                r"|([0-9A-Fa-f:.]+:[0-9A-Fa-f:.]*)"
                )
        RegularExpressionConversion.__init__(self, expr)

    def __call__(self, value):
        result = RegularExpressionConversion.__call__(self, value).lower()
        # Use C library to validate IPv6 addresses, in particular wrt.
        # number of colons and number of digits per group
        if ':' in result:
            import socket
            try:
                socket.inet_pton(socket.AF_INET6, result)
            except OSError:
                raise ValueError('%r is not a valid IPv6 address' % value)
        return result


def existing_directory(v):
    nv = os.path.expanduser(v)
    if os.path.isdir(nv):
        return nv
    raise ValueError('%s is not an existing directory' % v)


def existing_path(v):
    nv = os.path.expanduser(v)
    if os.path.exists(nv):
        return nv
    raise ValueError('%s is not an existing path' % v)


def existing_file(v):
    nv = os.path.expanduser(v)
    if os.path.exists(nv):
        return nv
    raise ValueError('%s is not an existing file' % v)


def existing_dirpath(v):
    nv = os.path.expanduser(v)
    dirname = os.path.dirname(nv)
    if not dirname:
        # relative pathname with no directory component
        return nv
    if os.path.isdir(dirname):
        return nv
    raise ValueError('The directory named as part of the path %s '
                     'does not exist.' % v)


class SuffixMultiplier:
    # d is a dictionary of suffixes to integer multipliers.  If no suffixes
    # match, default is the multiplier.  Matches are case insensitive.  Return
    # values are in the fundamental unit.
    def __init__(self, d, default=1):
        self._d = d
        self._default = default
        # all keys must be the same size

        def check(a, b):
            if len(a) != len(b):
                raise ValueError("suffix length mismatch")
            return a

        self._keysz = len(reduce(check, d))

    def __call__(self, v):
        v = v.lower()
        for s, m in self._d.items():
            if v[-self._keysz:] == s:
                return int(v[:-self._keysz]) * m
        return int(v) * self._default


def timedelta(s):
    # Unlike the standard time-interval data type, which returns a float
    # number of seconds, this datatype takes a wider range of syntax and
    # returns a datetime.timedelta
    #
    # Accepts suffixes:
    #    w - weeks
    #    d - days
    #    h - hours
    #    m - minutes
    #    s - seconds
    #
    # and all arguments may be integers or floats, positive or negative.
    # More than one time interval suffix value may appear on the line, but
    # they should all be separated by spaces, e.g.:
    #
    # sleep_time 4w 2d 7h 12m 0.00001s
    weeks = days = hours = minutes = seconds = 0
    for part in s.split():
        val = float(part[:-1])
        suffix = part[-1]
        if suffix == 'w':
            weeks = val
        elif suffix == 'd':
            days = val
        elif suffix == 'h':
            hours = val
        elif suffix == 'm':
            minutes = val
        elif suffix == 's':
            seconds = val
        else:
            raise TypeError('bad part {} in {}'.format(part, s))
    return datetime.timedelta(weeks=weeks, days=days, hours=hours,
                              minutes=minutes, seconds=seconds)


stock_datatypes = {
    "boolean":           asBoolean,
    "dotted-name":       DottedNameConversion(),
    "dotted-suffix":     DottedNameSuffixConversion(),
    "identifier":        IdentifierConversion(),
    "integer":           integer,
    "float":             float_conversion,
    "string":            str,
    "string-list":       string_list,
    "null":              null_conversion,
    "locale":            MemoizedConversion(check_locale),
    "port-number":       port_number,
    "basic-key":         BasicKeyConversion(),
    "inet-address":      inet_address,
    "inet-binding-address":    inet_binding_address,
    "inet-connection-address": inet_connection_address,
    "socket-address":    SocketAddress,
    "socket-binding-address":    SocketBindingAddress,
    "socket-connection-address": SocketConnectionAddress,
    "ipaddr-or-hostname": IpaddrOrHostname(),
    "existing-directory": existing_directory,
    "existing-path":     existing_path,
    "existing-file":     existing_file,
    "existing-dirpath":  existing_dirpath,
    "byte-size":         SuffixMultiplier({'kb': 1024,
                                           'mb': 1024*1024,
                                           'gb': 1024*1024*1024,
                                           }),
    "time-interval":     SuffixMultiplier({'s': 1,
                                           'm': 60,
                                           'h': 60*60,
                                           'd': 60*60*24,
                                           }),
    "timedelta":         timedelta,
    }


class Registry:
    """Implementation of a simple type registry.

    If given, *stock* should be a mapping which defines the "built-in"
    data types for the registry; if omitted or ``None``, the standard
    set of data types is used (see :ref:`standard-datatypes`).

    """

    def __init__(self, stock=None):
        if stock is None:
            stock = stock_datatypes.copy()
        self._stock = stock
        self._other = {}
        self._basic_key = None

    def find_name(self, conversion):
        """Return the best name for *conversion*, which must have been returned
        from *get* on this object."""
        for dct in self._other, self._stock:
            for k, v in dct.items():
                if v is conversion:
                    return k

        # If they followed the rules, we shouldn't get here.
        return str(conversion)  # pragma: no cover

    def get(self, name):
        """Return the type conversion routine for *name*.

        If the conversion function cannot be found, an (unspecified)
        exception is raised. If the name is not provided in the stock
        set of data types by this registry and has not otherwise been
        registered, this method uses the :meth:`search` method to load
        the conversion function. This is the only method the rest of
        :mod:`ZConfig` requires.

        """
        if '.' not in name:
            if self._basic_key is None:
                self._basic_key = self._other.get("basic-key")
                if self._basic_key is None:
                    self._basic_key = self._stock.get("basic-key")
                if self._basic_key is None:
                    self._basic_key = stock_datatypes["basic-key"]
            name = self._basic_key(name)
        t = self._stock.get(name)
        if t is None:
            t = self._other.get(name)
            if t is None:
                t = self.search(name)
        return t

    def register(self, name, conversion):
        """Register the data type name *name* to use the conversion function
        *conversion*.

        If *name* is already registered or provided as a stock data
        type, :exc:`ValueError` is raised (this includes the case when
        *name* was found using the :meth:`search` method).

        """
        if name in self._stock:
            raise ValueError("datatype name conflicts with built-in type: "
                             + repr(name))
        if name in self._other:
            raise ValueError("datatype name already registered: " + repr(name))
        self._other[name] = conversion

    def search(self, name):
        """This is a helper method for the default implementation of the
        :meth:`get` method.

        If *name* is a Python dotted-name, this method loads the value
        for the name by dynamically importing the containing module
        and extracting the value of the name. The name must refer to a
        usable conversion function.

        """
        if "." not in name:
            raise ValueError("unloadable datatype name: " + repr(name))
        components = name.split('.')
        start = components[0]
        g = {}
        package = __import__(start, g, g)
        modulenames = [start]
        for component in components[1:]:
            modulenames.append(component)
            try:
                package = getattr(package, component)
            except AttributeError:
                n = '.'.join(modulenames)
                package = __import__(n, g, g, component)
        self._other[name] = package
        return package



##############################################################################
#
# Copyright (c) 2001, 2018 Zope Foundation and Contributors.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################

"""Handlers which can plug into a PEP 282 logger."""

import logging
import logging.handlers
import os
import weakref
from sys import maxsize


# Export these, they're used in handlers.py
StreamHandler = logging.StreamHandler
SysLogHandler = logging.handlers.SysLogHandler
HTTPHandler = logging.handlers.HTTPHandler
SMTPHandler = logging.handlers.SMTPHandler
Win32EventLogHandler = logging.handlers.NTEventLogHandler


_reopenable_handlers = []


def closeFiles():
    """Reopen all logfiles managed by ZConfig configuration."""
    while _reopenable_handlers:
        wr = _reopenable_handlers.pop()
        h = wr()
        if h is not None:
            h.close()


def reopenFiles():
    """Reopen all logfiles managed by ZConfig configuration."""
    for wr in _reopenable_handlers[:]:
        h = wr()
        if h is None:
            try:
                _reopenable_handlers.remove(wr)
            except ValueError:
                continue
        else:
            h.reopen()


def _remove_from_reopenable(wr):
    try:
        _reopenable_handlers.remove(wr)
    except ValueError:
        pass


class FileHandler(logging.FileHandler):
    """File handler which supports reopening of logs.

    Re-opening should be used instead of the 'rollover' feature of
    the FileHandler from the standard library's logging package.
    """

    def __init__(self, filename, mode="a", encoding=None, delay=False):
        logging.FileHandler.__init__(self, filename,
                                     mode=mode, encoding=encoding, delay=delay)
        self._wr = weakref.ref(self, _remove_from_reopenable)
        _reopenable_handlers.append(self._wr)

    def close(self):
        # This can raise a KeyError if the handler has already been
        # removed, but a later error can be raised if
        # StreamHandler.close() isn't called.  This seems the best
        # compromise.  :-(
        try:
            logging.FileHandler.close(self)
        except KeyError:  # pragma: no cover
            pass
        _remove_from_reopenable(self._wr)

    def reopen(self):
        self.acquire()
        try:
            if self.stream is not None:
                self.stream.close()
                if self.delay:
                    self.stream = None
                else:
                    self.stream = self._open()
        finally:
            self.release()


class Win32FileHandler(FileHandler):
    """File-based log handler for Windows that supports an additional 'rotate'
    method.  reopen() is generally useless since Windows cannot do a move on
    an open file.

    """

    def rotate(self, rotateFilename=None):
        if not rotateFilename:
            rotateFilename = self.baseFilename + ".last"
        self.close()
        try:
            os.rename(self.baseFilename, rotateFilename)
        except OSError:
            pass

        if self.delay:
            self.stream = None
        else:
            self.stream = self._open()


if os.name == "nt":
    # Make it the default for Windows - we install a 'reopen' handler that
    # tries to rotate the logfile.
    FileHandler = Win32FileHandler


class RotatingFileHandler(logging.handlers.RotatingFileHandler):

    def __init__(self, *args, **kw):
        logging.handlers.RotatingFileHandler.__init__(self, *args, **kw)
        self._wr = weakref.ref(self, _remove_from_reopenable)
        _reopenable_handlers.append(self._wr)

    def close(self):
        logging.handlers.RotatingFileHandler.close(self)
        _remove_from_reopenable(self._wr)

    def reopen(self):
        self.doRollover()


class TimedRotatingFileHandler(logging.handlers.TimedRotatingFileHandler):

    def __init__(self, *args, **kw):
        logging.handlers.TimedRotatingFileHandler.__init__(self, *args, **kw)
        self._wr = weakref.ref(self, _remove_from_reopenable)
        _reopenable_handlers.append(self._wr)

    def close(self):
        logging.handlers.TimedRotatingFileHandler.close(self)
        _remove_from_reopenable(self._wr)

    def reopen(self):
        self.doRollover()


class NullHandler(logging.NullHandler):
    """Handler that does nothing."""


class StartupHandler(logging.handlers.BufferingHandler):
    """Handler which stores messages in a buffer until later.

    This is useful at startup before we can know that we can safely
    write to a configuration-specified handler.
    """

    def __init__(self):
        logging.handlers.BufferingHandler.__init__(self, maxsize)

    def shouldFlush(self, record):
        return False

    def flushBufferTo(self, target):
        while self.buffer:
            target.handle(self.buffer.pop(0))



##############################################################################
#
# Copyright (c) 2017, 2018 Zope Foundation and Contributors.
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################

import unittest

from ZConfig import ConfigurationError
from ZConfig import SchemaError
from ZConfig.info import AbstractType
from ZConfig.info import BaseInfo
from ZConfig.info import BaseKeyInfo
from ZConfig.info import KeyInfo
from ZConfig.info import SchemaType
from ZConfig.info import SectionInfo
from ZConfig.info import SectionType
from ZConfig.info import Unbounded
from ZConfig.tests.support import TestHelper


class UnboundTestCase(unittest.TestCase):

    def test_order(self):
        self.assertGreater(Unbounded, self)
        self.assertFalse(Unbounded > Unbounded)
        self.assertEqual(Unbounded, Unbounded)

    def test_repr(self):
        self.assertEqual(repr(Unbounded), '<Unbounded>')


class InfoMixin(TestHelper):

    Class = None

    default_kwargs = {'name': '', 'datatype': None, 'handler': None,
                      'minOccurs': 0, 'maxOccurs': Unbounded,
                      'attribute': None}

    def make_one(self, **kwargs):
        args = self.default_kwargs.copy()
        args.update(kwargs)
        return self.Class(**args)


class BaseInfoTestCase(InfoMixin, unittest.TestCase):

    Class = BaseInfo

    def test_constructor_error(self):
        self.assertRaisesRegex(SchemaError,
                               'maxOccurs',
                               self.make_one,
                               maxOccurs=0, minOccurs=0)

        # This case doesn't really make sense
        self.assertRaisesRegex(SchemaError,
                               'minOccurs',
                               self.make_one,
                               maxOccurs=1,
                               minOccurs=2)

    def test_repr(self):
        # just doesn't raise
        repr(self.make_one())


class BaseKeyInfoTestCase(InfoMixin, unittest.TestCase):

    class Class(BaseKeyInfo):
        def add_valueinfo(self, vi, key):
            "This wont actually be called"

    def test_cant_instantiate(self):
        self.Class = BaseKeyInfo
        with self.assertRaises(TypeError):
            self.make_one()
        del self.Class

    def test_finish(self):
        info = self.make_one(minOccurs=1)
        info.finish()
        with self.assertRaises(SchemaError):
            info.finish()

    def test_adddefaultc(self):
        info = self.make_one(name='foo', minOccurs=1)
        self.assertRaisesRegex(SchemaError,
                               'unexpected key for default',
                               info.adddefault,
                               None, None, key='key')


class KeyInfoTestCase(InfoMixin, unittest.TestCase):

    Class = KeyInfo
    default_kwargs = InfoMixin.default_kwargs.copy()
    default_kwargs.pop('maxOccurs')

    def test_add_with_default(self):
        info = self.make_one(minOccurs=1, name='name')
        info.adddefault('value', None)
        self.assertRaisesRegex(SchemaError,
                               'cannot set more than one',
                               info.adddefault,
                               'value', None)


class SectionInfoTestCase(InfoMixin, unittest.TestCase):

    Class = SectionInfo

    class MockSectionType:
        name = None

        @classmethod
        def isabstract(cls):
            return True

    default_kwargs = InfoMixin.default_kwargs.copy()
    default_kwargs.pop('datatype')
    default_kwargs['sectiontype'] = MockSectionType

    def test_constructor_error(self):
        self.assertRaisesRegex(SchemaError,
                               'must use a name',
                               self.make_one,
                               name='name', maxOccurs=2)
        self.assertRaisesRegex(SchemaError,
                               'must specify a target attribute',
                               self.make_one,
                               name='*', maxOccurs=2)

    def test_misc(self):
        info = self.make_one(maxOccurs=1)
        repr(info)
        self.assertFalse(info.isAllowedName('*'))
        self.assertFalse(info.isAllowedName('+'))


class AbstractTypeTestCase(unittest.TestCase):

    def test_subtypes(self):

        t = AbstractType('name')
        self.assertFalse(t.hassubtype('foo'))
        self.assertEqual([], list(t.getsubtypenames()))

        self.name = 'foo'
        t.addsubtype(self)
        self.assertTrue(t.hassubtype('foo'))


class SectionTypeTestCase(TestHelper, unittest.TestCase):

    def make_one(self, name='', keytype=None, valuetype=None,
                 datatype=None, registry={}, types=None):
        return SectionType(name, keytype, valuetype, datatype, registry, types)

    def test_getinfo_no_key(self):
        info = self.make_one()
        self.assertRaisesRegex(ConfigurationError,
                               "cannot match a key without a name",
                               info.getinfo,
                               None)

    def test_required_types_with_name(self):
        info = self.make_one(name='foo')
        self.assertEqual(['foo'], info.getrequiredtypes())

    def test_getsectioninfo(self):

        class MockChild:
            _issection = False

            def issection(self):
                return self._issection

        child = MockChild()

        info = self.make_one()

        info._children.append(('foo', child))

        self.assertRaisesRegex(ConfigurationError,
                               'already in use for key',
                               info.getsectioninfo,
                               None, 'foo')

        self.assertRaisesRegex(ConfigurationError,
                               'no matching section',
                               info.getsectioninfo,
                               None, 'baz')


class SchemaTypeTestCase(TestHelper, unittest.TestCase):

    def test_various(self):
        class Mock:
            pass

        mock = Mock()
        schema = SchemaType(None, None, None, None, 'url', {})

        mock.name = 'name'
        schema.addtype(mock)
        with self.assertRaises(SchemaError):
            schema.addtype(mock)

        self.assertTrue(schema.allowUnnamed())
        self.assertFalse(schema.isAllowedName(None))

        with self.assertRaises(SchemaError):
            schema.deriveSectionType(schema, None, None, None, None)

        schema.addComponent('name')
        self.assertRaisesRegex(SchemaError,
                               'already have component',
                               schema.addComponent,
                               'name')



##############################################################################
#
# Copyright (c) 2017, 2018 Zope Foundation and Contributors.
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################

import unittest

from ZConfig import ConfigurationError
from ZConfig import DataConversionError
from ZConfig.matcher import BaseMatcher
from ZConfig.matcher import SectionMatcher
from ZConfig.matcher import SectionValue
from ZConfig.tests.support import TestHelper


class SectionValueTestCase(unittest.TestCase):

    def test_repr(self):

        class MockMatcher:
            type = None

        matcher = MockMatcher()
        matcher.type = MockMatcher()
        matcher.type.name = 'matcher'

        sv = SectionValue({}, 'name', matcher)
        self.assertIn('name', repr(sv))

        sv = SectionValue({}, None, matcher)
        self.assertIn('at', repr(sv))

        self.assertIs(matcher, sv.getSectionMatcher())

    def test_str(self):
        d = {'k': 'v'}
        sv = SectionValue(d, None, None)
        self.assertEqual(
            'k                                       : v',
            str(sv))


class SectionMatcherTestCase(TestHelper, unittest.TestCase):

    def test_constructor_error(self):

        class Mock:
            name = 'name'

            def allowUnnamed(self):
                return False

        mock = Mock()
        self.assertRaisesRegex(ConfigurationError,
                               "sections may not be unnamed",
                               SectionMatcher,
                               mock, mock, None, None)


class BaseMatcherTestCase(TestHelper, unittest.TestCase):

    def test_repr(self):

        class Mock(dict):
            name = 'name'

        matcher = BaseMatcher(None, Mock(), None)
        repr(matcher)

    def test_duplicate_section_names(self):

        class Mock(dict):
            name = 'name'

        matcher = BaseMatcher(None, Mock(), None)
        matcher._sectionnames['foo'] = None

        self.assertRaisesRegex(ConfigurationError,
                               "section names must not be re-used",
                               matcher.addSection,
                               None, 'foo', None)

    def test_construct_errors(self):

        class MockType:
            attribute = 'attr'

            _multi = True
            _section = True

            def ismulti(self):
                return self._multi

            def issection(self):
                return self._section

        type_ = []
        matcher = BaseMatcher(None, type_, None)
        type_.append(('key', MockType()))

        class MockSection:
            def getSectionDefinition(self):
                return self

            def datatype(self, _s):
                raise ValueError()

        matcher._values['attr'] = [MockSection()]

        with self.assertRaises(DataConversionError):
            matcher.constuct()

        type_[0][1]._multi = False
        matcher._values['attr'] = MockSection()
        with self.assertRaises(DataConversionError):
            matcher.constuct()

    def test_create_child_bad_name(self):

        class MockType(list):
            name = 'foo'
            sectiontype = None

            def getsectioninfo(self, type_name, name):
                return self

            def isabstract(self):
                return False

            def isAllowedName(self, name):
                return False

        t = MockType()
        t.sectiontype = MockType()
        matcher = BaseMatcher(None, t, None)
        self.assertRaisesRegex(ConfigurationError,
                               'is not an allowed name',
                               matcher.createChildMatcher,
                               MockType(), 'ignored')



##############################################################################
#
# Copyright (c) 2002, 2003, 2018 Zope Foundation and Contributors.
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################
"""Tests of ZConfig.loader classes and helper functions."""

import os.path
import sys
import tempfile
import unittest
import urllib.request
from io import StringIO

import ZConfig
import ZConfig.loader
import ZConfig.url
from ZConfig.tests.support import CONFIG_BASE
from ZConfig.tests.support import TestHelper


myfile = os.path.abspath(__file__)
LIBRARY_DIR = os.path.join(os.path.dirname(myfile), "library")


class LoaderTestCase(TestHelper, unittest.TestCase):

    def test_open_resource_non_ascii(self):
        # Files are decoded using utf-8 on open
        loader = ZConfig.loader.SchemaLoader()
        url = ZConfig.url.urljoin(CONFIG_BASE, "non-ascii.txt")
        with loader.openResource(url) as stream:
            val = stream.read()
        self.assertEqual(
            val,
            '# -*-coding: utf-8; mode: conf-*-\n'
            'This file contains a snowman, U+2603: \u2603\n'
        )

    def test_schema_caching(self):
        loader = ZConfig.loader.SchemaLoader()
        url = ZConfig.url.urljoin(CONFIG_BASE, "simple.xml")
        schema1 = loader.loadURL(url)
        schema2 = loader.loadURL(url)
        self.assertIs(schema1, schema2)

    def test_simple_import_with_cache(self):
        loader = ZConfig.loader.SchemaLoader()
        url1 = ZConfig.url.urljoin(CONFIG_BASE, "library.xml")
        schema1 = loader.loadURL(url1)
        sio = StringIO("<schema>"
                       "  <import src='library.xml'/>"
                       "  <section type='type-a' name='section'/>"
                       "</schema>")
        url2 = ZConfig.url.urljoin(CONFIG_BASE, "stringio")
        schema2 = loader.loadFile(sio, url2)
        self.assertTrue(schema1.gettype("type-a") is schema2.gettype("type-a"))

    def test_schema_loader_source_errors(self):
        loader = ZConfig.loader.SchemaLoader()
        self.assertRaisesRegex(ZConfig.SchemaError,
                               "illegal schema component name",
                               loader.schemaComponentSource,
                               '', None)
        self.assertRaisesRegex(ZConfig.SchemaError,
                               "illegal schema component name",
                               loader.schemaComponentSource,
                               'foo..bar', None)

    def test_config_loader_abstract_schema(self):

        class MockSchema:
            _abstract = True

            def isabstract(self):
                return self._abstract

            def gettype(self, _t):
                return self

        self.assertRaisesRegex(ZConfig.SchemaError,
                               "abstract type",
                               ZConfig.loader.ConfigLoader,
                               MockSchema())

        s = MockSchema()
        s._abstract = False

        loader = ZConfig.loader.ConfigLoader(s)
        s._abstract = True

        self.assertRaisesRegex(ZConfig.ConfigurationError,
                               "cannot match abstract section",
                               loader.startSection,
                               None, None, None)

    def test_simple_import_using_prefix(self):
        self.load_schema_text("""\
            <schema prefix='ZConfig.tests.library'>
              <import package='.thing'/>
            </schema>
            """)

    def test_import_errors(self):
        # must specify exactly one of package or src
        self.assertRaises(ZConfig.SchemaError, ZConfig.loadSchemaFile,
                          StringIO("<schema><import/></schema>"))
        self.assertRaises(ZConfig.SchemaError, ZConfig.loadSchemaFile,
                          StringIO("<schema>"
                                   "  <import src='library.xml'"
                                   "          package='ZConfig'/>"
                                   "</schema>"))
        # cannot specify src and file
        self.assertRaises(ZConfig.SchemaError, ZConfig.loadSchemaFile,
                          StringIO("<schema>"
                                   "  <import src='library.xml'"
                                   "          file='other.xml'/>"
                                   "</schema>"))
        # cannot specify module as package
        sio = StringIO("<schema>"
                       "  <import package='ZConfig.tests.test_loader'/>"
                       "</schema>")
        with self.assertRaises(ZConfig.SchemaResourceError) as ctx:
            ZConfig.loadSchemaFile(sio)

        e = ctx.exception
        self.assertEqual(e.filename, "component.xml")
        self.assertEqual(e.package, "ZConfig.tests.test_loader")
        self.assertTrue(e.path is None)
        # make sure the str() doesn't raise an unexpected exception
        str(e)

    def test_import_from_package(self):
        loader = ZConfig.loader.SchemaLoader()
        sio = StringIO("<schema>"
                       "  <import package='ZConfig.tests.library.widget'/>"
                       "</schema>")
        schema = loader.loadFile(sio)
        self.assertTrue(schema.gettype("widget-a") is not None)

    def test_import_from_package_with_file(self):
        loader = ZConfig.loader.SchemaLoader()
        sio = StringIO("<schema>"
                       "  <import package='ZConfig.tests.library.widget'"
                       "          file='extra.xml' />"
                       "</schema>")
        schema = loader.loadFile(sio)
        self.assertTrue(schema.gettype("extra-type") is not None)

    def test_import_from_package_extra_directory(self):
        loader = ZConfig.loader.SchemaLoader()
        sio = StringIO("<schema>"
                       "  <import package='ZConfig.tests.library.thing'"
                       "          file='extras.xml' />"
                       "</schema>")
        schema = loader.loadFile(sio)
        self.assertTrue(schema.gettype("extra-thing") is not None)

    def test_import_from_package_with_missing_file(self):
        loader = ZConfig.loader.SchemaLoader()
        sio = StringIO("<schema>"
                       "  <import package='ZConfig.tests.library.widget'"
                       "          file='notthere.xml' />"
                       "</schema>")
        with self.assertRaises(ZConfig.SchemaResourceError) as ctx:
            loader.loadFile(sio)
        e = ctx.exception
        self.assertEqual(e.filename, "notthere.xml")
        self.assertEqual(e.package, "ZConfig.tests.library.widget")
        self.assertTrue(e.path)
        # make sure the str() doesn't raise an unexpected exception
        str(e)

    def test_import_from_package_with_directory_file(self):
        loader = ZConfig.loader.SchemaLoader()
        sio = StringIO("<schema>"
                       "  <import package='ZConfig.tests.library.widget'"
                       "          file='really/notthere.xml' />"
                       "</schema>")
        self.assertRaises(ZConfig.SchemaError, loader.loadFile, sio)

    def test_import_two_components_one_package(self):
        loader = ZConfig.loader.SchemaLoader()
        sio = StringIO("<schema>"
                       "  <import package='ZConfig.tests.library.widget' />"
                       "  <import package='ZConfig.tests.library.widget'"
                       "          file='extra.xml' />"
                       "</schema>")
        schema = loader.loadFile(sio)
        schema.gettype("widget-a")
        schema.gettype("extra-type")

    def test_import_component_twice_1(self):
        # Make sure we can import a component twice from a schema.
        # This is most likely to occur when the component is imported
        # from each of two other components, or from the top-level
        # schema and a component.
        loader = ZConfig.loader.SchemaLoader()
        sio = StringIO("<schema>"
                       "  <import package='ZConfig.tests.library.widget' />"
                       "  <import package='ZConfig.tests.library.widget' />"
                       "</schema>")
        schema = loader.loadFile(sio)
        schema.gettype("widget-a")

    def test_import_component_twice_2(self):
        # Make sure we can import a component from a config file even
        # if it has already been imported from the schema.
        loader = ZConfig.loader.SchemaLoader()
        sio = StringIO("<schema>"
                       "  <import package='ZConfig.tests.library.widget' />"
                       "</schema>")
        schema = loader.loadFile(sio)
        loader = ZConfig.loader.ConfigLoader(schema)
        sio = StringIO("%import ZConfig.tests.library.widget")
        loader.loadFile(sio)

    def test_file_url_normalization(self):
        self.assertEqual(
            ZConfig.url.urlnormalize("file:/abc/def"),
            "file:///abc/def")
        self.assertEqual(
            ZConfig.url.urlunsplit(("file", "", "/abc/def", "", "")),
            "file:///abc/def")
        self.assertEqual(
            ZConfig.url.urljoin("file:/abc/", "def"),
            "file:///abc/def")
        self.assertEqual(
            ZConfig.url.urldefrag("file:/abc/def#frag"),
            ("file:///abc/def", "frag"))

    def test_url_from_file(self):
        class MockFile:
            name = 'path'
        self.assertEqual('file://',
                         ZConfig.loader._url_from_file(MockFile)[:7])

    def test_isPath(self):
        assertTrue = self.assertTrue
        isPath = ZConfig.loader.SchemaLoader().isPath
        assertTrue(isPath("abc"))
        assertTrue(isPath("abc/def"))
        assertTrue(isPath("/abc"))
        assertTrue(isPath("/abc/def"))
        assertTrue(isPath(r"\abc"))
        assertTrue(isPath(r"\abc\def"))
        assertTrue(isPath(r"c:\abc\def"))
        assertTrue(isPath("/ab:cd"))
        assertTrue(isPath(r"\ab:cd"))
        assertTrue(isPath("long name with spaces"))
        assertTrue(isPath("long name:with spaces"))
        assertTrue(not isPath("ab:cd"))
        assertTrue(not isPath("http://www.example.com/"))
        assertTrue(not isPath("http://www.example.com/sample.conf"))
        assertTrue(not isPath("file:///etc/zope/zope.conf"))
        assertTrue(not isPath("file:///c|/foo/bar.conf"))


class TestNonExistentResources(unittest.TestCase):

    # XXX Not sure if this is the best approach for these.  These
    # tests make sure that the error reported by ZConfig for missing
    # resources is handled in a consistent way.  Since ZConfig uses
    # urllib2.urlopen() for opening all resources, what we do is
    # replace that function with one that always raises an exception.
    # Since urllib2.urlopen() can raise either IOError or OSError
    # (depending on the version of Python), we run test for each
    # exception.  urllib2.urlopen() is restored after running the
    # test.

    def setUp(self):
        self.urllib2_urlopen = urllib.request.urlopen
        urllib.request.urlopen = self.fake_urlopen

    def tearDown(self):
        urllib.request.urlopen = self.urllib2_urlopen

    def fake_urlopen(self, url):
        raise self.error()

    def test_nonexistent_file_ioerror(self):
        self.error = IOError
        self.check_nonexistent_file()

    def test_nonexistent_file_oserror(self):
        self.error = OSError
        self.check_nonexistent_file()

    def check_nonexistent_file(self):
        fn = tempfile.mktemp()
        schema = ZConfig.loadSchemaFile(StringIO("<schema/>"))
        self.assertRaises(ZConfig.ConfigurationError,
                          ZConfig.loadSchema, fn)
        self.assertRaises(ZConfig.ConfigurationError,
                          ZConfig.loadConfig, schema, fn)
        self.assertRaises(ZConfig.ConfigurationError,
                          ZConfig.loadConfigFile, schema,
                          StringIO("%include " + fn))
        self.assertRaises(ZConfig.ConfigurationError,
                          ZConfig.loadSchema,
                          "http://www.zope.org/no-such-document/")
        self.assertRaises(ZConfig.ConfigurationError,
                          ZConfig.loadConfig, schema,
                          "http://www.zope.org/no-such-document/")


class TestResourcesInZip(unittest.TestCase):

    def setUp(self):
        self.old_path = sys.path[:]
        # now add our sample EGG to sys.path:
        zipfile = os.path.join(os.path.dirname(myfile), "foosample.zip")
        sys.path.append(zipfile)

    def tearDown(self):
        sys.path[:] = self.old_path

    def test_zip_import_component_from_schema(self):
        sio = StringIO('''
            <schema>
              <abstracttype name="something"/>
              <import package="foo.sample"/>
              <section name="*"
                       attribute="something"
                       type="something"
                       />
            </schema>
            ''')
        schema = ZConfig.loadSchemaFile(sio)
        t = schema.gettype("sample")
        self.assertFalse(t.isabstract())

    def test_zip_import_component_from_config(self):
        sio = StringIO('''
            <schema>
              <abstracttype name="something"/>
              <section name="*"
                       attribute="something"
                       type="something"
                       />
            </schema>
            ''')
        schema = ZConfig.loadSchemaFile(sio)

        value = '''
            %import foo.sample
            <sample>
              data value
            </sample>
            '''
        sio = StringIO(value)
        config, _ = ZConfig.loadConfigFile(schema, sio)
        self.assertEqual(config.something.data, "| value |")

        sio = StringIO(value)
        with self.assertRaises(ZConfig.ConfigurationSyntaxError):
            ZConfig.loadConfigFile(schema, sio,
                                   overrides=["sample/data=othervalue"])


class TestOpenPackageResource(TestHelper, unittest.TestCase):

    magic_name = 'not a valid import name'

    def setUp(self):
        sys.modules[self.magic_name] = self

    def tearDown(self):
        del sys.modules[self.magic_name]

    def test_package_loader_resource_error(self):
        class MockLoader:
            pass
        self.__loader__ = MockLoader()
        self.__path__ = ['dir']

        self.assertRaisesRegex(ZConfig.SchemaResourceError,
                               "error opening schema component",
                               ZConfig.loader.openPackageResource,
                               self.magic_name, 'a path')

        # Now with an empty path
        self.__path__ = []
        self.assertRaisesRegex(ZConfig.SchemaResourceError,
                               "schema component not found",
                               ZConfig.loader.openPackageResource,
                               self.magic_name, 'a path')

    def test_resource(self):
        r = ZConfig.loader.Resource(self, None)
        self.assertEqual(self.magic_name, r.magic_name)



##############################################################################
#
# Copyright (c) 2002, 2003, 2018 Zope Foundation and Contributors.
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################
"""Schema loader utility."""

import os.path
import re
import sys
import urllib.request
from abc import ABC
from abc import abstractmethod
from io import StringIO
from urllib.request import pathname2url

import ZConfig
import ZConfig.cfgparser
import ZConfig.datatypes
import ZConfig.info
import ZConfig.matcher
import ZConfig.schema
import ZConfig.url


def loadSchema(url):
    """Load a schema definition from the URL *url*.

    *url* may be a URL, absolute pathname, or relative pathname.
    Fragment identifiers are not supported.

    The resulting schema object can be passed to :func:`loadConfig` or
    :func:`loadConfigFile`. The schema object may be used as many
    times as needed.

    .. seealso:: :class:`~.SchemaLoader`, :meth:`.BaseLoader.loadURL`
    """
    return SchemaLoader().loadURL(url)


def loadSchemaFile(file, url=None):
    """Load a schema definition from the open file object *file*.

    If *url* is given and not ``None``, it should be the URL of
    resource represented by *file*. If *url* is omitted or ``None``, a
    URL may be computed from the ``name`` attribute of *file*, if
    present. The resulting schema object can be passed to
    :func:`loadConfig` or :func:`loadConfigFile`. The schema object
    may be used as many times as needed.

    .. seealso:: :class:`~.SchemaLoader`, :meth:`.BaseLoader.loadFile`
    """
    return SchemaLoader().loadFile(file, url)


def loadConfig(schema, url, overrides=()):
    """Load and return a configuration from a URL or pathname given by
    *url*.

    *url* may be a URL, absolute pathname, or relative pathname.
    Fragment identifiers are not supported. *schema* is a reference to a
    schema loaded by :func:`loadSchema` or :func:`loadSchemaFile`.

    The return value is a tuple containing the configuration object and
    a composite handler that, when called with a name-to-handler
    mapping, calls all the handlers for the configuration.

    The optional *overrides* argument represents information derived
    from command-line arguments. If given, it must be either a
    sequence of value specifiers, or ``None``. A "value specifier" is
    a string of the form ``optionpath=value``, for example,
    ``some/path/to/key=value``.

    .. seealso::
       :meth:`.ExtendedConfigLoader.addOption`
            For information on the format of value specifiers.
       :class:`~.ConfigLoader`
            For information about loading configs.
       :meth:`.BaseLoader.loadURL`
            For information about the format of *url*
    """
    return _get_config_loader(schema, overrides).loadURL(url)


def loadConfigFile(schema, file, url=None, overrides=()):
    """Load and return a configuration from an opened file object.

    If *url* is omitted, one will be computed based on the ``name``
    attribute of *file*, if it exists. If no URL can be determined,
    all ``%include`` statements in the configuration must use absolute
    URLs. *schema* is a reference to a schema loaded by
    :func:`loadSchema` or :func:`loadSchemaFile`.

    The return value is a tuple containing the configuration object
    and a composite handler that, when called with a name-to-handler
    mapping, calls all the handlers for the configuration. The
    *overrides* argument is the same as for the :func:`loadConfig`
    function.

    .. seealso:: :class:`~.ConfigLoader`, :meth:`.BaseLoader.loadFile`,
       :meth:`.ExtendedConfigLoader.addOption`
    """
    return _get_config_loader(schema, overrides).loadFile(file, url)


def _get_config_loader(schema, overrides):
    if overrides:
        from ZConfig import cmdline
        loader = cmdline.ExtendedConfigLoader(schema)
        for opt in overrides:
            loader.addOption(opt)
    else:
        loader = ConfigLoader(schema)
    return loader


class BaseLoader(ABC):
    """Base class for loader objects.

    This should not be instantiated
    directly, as the :meth:`loadResource` method must be overridden
    for the instance to be used via the public API.
    """

    def __init__(self):
        pass

    def createResource(self, file, url):
        """Returns a resource object for an open file and URL, given as *file*
        and *url*, respectively.

        This may be overridden by a subclass if an alternate resource
        implementation is desired.
        """
        return Resource(file, url)

    def loadURL(self, url):
        """Open and load a resource specified by the URL *url*.

        This method uses the :meth:`loadResource` method to perform the
        actual load, and returns whatever that method returns.
        """
        url = self.normalizeURL(url)
        with self.openResource(url) as r:
            return self.loadResource(r)

    def loadFile(self, file, url=None):
        """Load from an open file object, *file*.

        If given and not ``None``, *url* should be the URL of the
        resource represented by *file*. If omitted or *None*, the
        ``name`` attribute of *file* is used to compute a ``file:``
        URL, if present.

        This method uses the :meth:`loadResource` method to perform the
        actual load, and returns whatever that method returns.
        """
        if not url:
            url = _url_from_file(file)
        with self.createResource(file, url) as r:
            return self.loadResource(r)

    # utilities

    @abstractmethod
    def loadResource(self, resource):
        """Abstract method.

        Subclasses of :class:`BaseLoader` must implement this method to
        actually load the resource and return the appropriate
        application-level object.
        """

    def openResource(self, url):
        """Returns a resource object that represents the URL *url*.

        The URL is opened using the :func:`urllib.request.urlopen` function,
        and the returned resource object is created using
        :meth:`createResource`. If the URL cannot be opened,
        :exc:`~.ConfigurationError` is raised.
        """
        # ConfigurationError exceptions raised here should be
        # str()able to generate a message for an end user.
        #
        # XXX This should be replaced to use a local cache for remote
        # resources.  The policy needs to support both re-retrieve on
        # change and provide the cached resource when the remote
        # resource is not accessible.
        url = str(url)
        if url.startswith("package:"):
            _, package, filename = url.split(":", 2)
            file = openPackageResource(package, filename)
        else:
            try:
                file = urllib.request.urlopen(url)
            except urllib.request.URLError as e:
                # urllib.request.URLError has a particularly hostile str(), so
                # we generally don't want to pass it along to the user.
                self._raise_open_error(url, e.reason)  # pragma: no cover
            except OSError as e:
                self._raise_open_error(url, str(e))

            try:
                data = file.read()
            finally:
                file.close()
            if isinstance(data, bytes):
                # Be sure to specify an (useful) encoding so we don't get
                # the system default, typically ascii.
                data = data.decode('utf-8')
            file = StringIO(data)
        return self.createResource(file, url)

    def _raise_open_error(self, url, message):
        if url[:7].lower() == "file://":
            what = "file"
            ident = urllib.request.url2pathname(url[7:])
        else:
            what = "URL"
            ident = url
        raise ZConfig.ConfigurationError(
            "error opening {} {}: {}".format(what, ident, message), url)

    def normalizeURL(self, url):
        """Return a URL for *url*

        If *url* refers to an existing file, the corresponding
        ``file:`` URL is returned. Otherwise *url* is checked
        for sanity: if it does not have a schema, :exc:`ValueError` is
        raised, and if it does have a fragment identifier,
        :exc:`~.ConfigurationError` is raised.

        This uses :meth:`isPath` to determine whether *url* is
        a URL of a filesystem path.
        """
        if self.isPath(url):
            url = "file://" + pathname2url(os.path.abspath(url))
        newurl, fragment = ZConfig.url.urldefrag(url)
        if fragment:
            raise ZConfig.ConfigurationError(
                "fragment identifiers are not supported",
                url)
        return newurl

    # from RFC 3986:
    # schema = ALPHA *( ALPHA / DIGIT / "+" / "-" / "." )
    _pathsep_rx = re.compile(r"[a-zA-Z][-+.a-zA-Z0-9]*:")

    def isPath(self, s):
        """Return true if *s* should be considered a filesystem path rather
        than a URL.
        """
        if ":" in s:
            # XXX This assumes that one-character scheme identifiers
            # are always Windows drive letters; I don't know of any
            # one-character scheme identifiers.
            m = self._pathsep_rx.match(s)
            if m is None:
                return True
            # Does it look like a drive letter?
            return len(m.group(0)) == 2
        else:
            return True


def openPackageResource(package, path):
    __import__(package)
    pkg = sys.modules[package]
    try:
        loader = pkg.__loader__
    except AttributeError:
        relpath = os.path.join(*path.split("/"))
        for dirname in pkg.__path__:
            filename = os.path.join(dirname, relpath)
            if os.path.exists(filename):
                break
        else:
            raise ZConfig.SchemaResourceError("schema component not found",
                                              filename=path,
                                              package=package,
                                              path=pkg.__path__)
        url = "file:" + pathname2url(filename)
        url = ZConfig.url.urlnormalize(url)
        return urllib.parse.urlopen(url)
    else:
        v, tb = (None, None)
        for dirname in pkg.__path__:
            loadpath = os.path.join(dirname, path)
            try:
                return StringIO(
                    loader.get_data(loadpath).decode('utf-8'))
            except Exception as e:
                v = ZConfig.SchemaResourceError(
                    "error opening schema component: " + repr(e),
                    filename=path,
                    package=package,
                    path=pkg.__path__)
                tb = sys.exc_info()[2]

        if v is not None:
            try:
                raise v.with_traceback(tb)
            finally:
                del tb

        raise ZConfig.SchemaResourceError("schema component not found",
                                          filename=path,
                                          package=package,
                                          path=pkg.__path__)


def _url_from_file(file_or_path):
    name = getattr(file_or_path, "name", None)
    if name and name[0] != "<" and name[-1] != ">":
        return "file://" + pathname2url(os.path.abspath(name))


class SchemaLoader(BaseLoader):
    """ Loader that loads schema instances.

    All schema loaded by a :class:`SchemaLoader` will use the same
    data type registry. If *registry* is provided and not ``None``, it
    will be used, otherwise an instance of
    :class:`ZConfig.datatypes.Registry` will be used.
    """

    def __init__(self, registry=None):
        if registry is None:
            registry = ZConfig.datatypes.Registry()
        BaseLoader.__init__(self)
        self.registry = registry
        self._cache = {}

    def loadResource(self, resource):
        if resource.url and resource.url in self._cache:
            schema = self._cache[resource.url]
        else:
            schema = ZConfig.schema.parseResource(resource, self)
            self._cache[resource.url] = schema
        return schema

    # schema parser support API

    def schemaComponentSource(self, package, filename):
        parts = package.split(".")
        if not parts:  # pragma: no cover
            raise ZConfig.SchemaError(
                "illegal schema component name: " + repr(package))
        if "" in parts:
            # '' somewhere in the package spec; still illegal
            raise ZConfig.SchemaError(
                "illegal schema component name: " + repr(package))
        filename = filename or "component.xml"
        try:
            __import__(package)
        except ImportError as e:
            raise ZConfig.SchemaResourceError(
                "could not load package {}: {}".format(package, str(e)),
                filename=filename,
                package=package)
        pkg = sys.modules[package]
        if not hasattr(pkg, "__path__"):
            raise ZConfig.SchemaResourceError(
                "import name does not refer to a package",
                filename=filename, package=package)
        return "package:{}:{}".format(package, filename)


class ConfigLoader(BaseLoader):
    """Loader for configuration files.

    Each configuration file must
    conform to the schema *schema*.  The ``load*()`` methods
    return a tuple consisting of the configuration object and a
    composite handler.

    """

    def __init__(self, schema):
        if schema.isabstract():
            raise ZConfig.SchemaError(
                "cannot check a configuration an abstract type")
        BaseLoader.__init__(self)
        self.schema = schema
        self._private_schema = False

    def loadResource(self, resource):
        sm = self.createSchemaMatcher()
        self._parse_resource(sm, resource)
        result = sm.finish(), CompositeHandler(sm.handlers, self.schema)
        return result

    def createSchemaMatcher(self):
        return ZConfig.matcher.SchemaMatcher(self.schema)

    # config parser support API

    def startSection(self, parent, type_, name):
        t = self.schema.gettype(type_)
        if t.isabstract():
            raise ZConfig.ConfigurationError(
                "concrete sections cannot match abstract section types;"
                " found abstract type " + repr(type_))
        return parent.createChildMatcher(t, name)

    def endSection(self, parent, type_, name, matcher):
        sectvalue = matcher.finish()
        parent.addSection(type_, name, sectvalue)

    def importSchemaComponent(self, pkgname):
        schema = self.schema
        if not self._private_schema:
            # replace the schema with an extended schema on the first %import
            self._loader = SchemaLoader(self.schema.registry)
            schema = ZConfig.info.createDerivedSchema(self.schema)
            self._private_schema = True
            self.schema = schema
        url = self._loader.schemaComponentSource(pkgname, '')
        if schema.hasComponent(url):
            return
        schema.addComponent(url)
        with self.openResource(url) as resource:
            ZConfig.schema.parseComponent(resource, self._loader, schema)

    def includeConfiguration(self, section, url, defines):
        url = self.normalizeURL(url)
        with self.openResource(url) as r:
            self._parse_resource(section, r, defines)

    # internal helper

    def _parse_resource(self, matcher, resource, defines=None):
        parser = ZConfig.cfgparser.ZConfigParser(resource, self, defines)
        parser.parse(matcher)


class CompositeHandler:

    def __init__(self, handlers, schema):
        self._handlers = handlers
        self._convert = schema.registry.get("basic-key")

    def __call__(self, handlermap):
        d = {}
        for name, callback in handlermap.items():
            n = self._convert(name)
            if n in d:
                raise ZConfig.ConfigurationError(
                    "handler name not unique when converted to a basic-key: "
                    + repr(name))
            d[n] = callback
        L = []
        for handler, value in self._handlers:
            if handler not in d:
                L.append(handler)
        if L:
            raise ZConfig.ConfigurationError(
                "undefined handlers: " + ", ".join(L))
        for handler, value in self._handlers:
            f = d[handler]
            if f is not None:
                f(value)

    def __len__(self):
        return len(self._handlers)


class Resource:
    """Object that allows an open file object and a URL to be bound
    together to ease handling.

    Instances have the attributes :attr:`file` and :attr:`url`, which
    store the constructor arguments. These objects also have a
    :meth:`close` method which will call :meth:`~file.close` on
    *file*, then set the :attr:`file` attribute to ``None`` and the
    :attr:`closed` attribute to ``True``. Using this object as a
    context manager also closes the file.

    All other attributes are delegated to *file*.
    """

    closed = False

    def __init__(self, file, url):
        self.file = file
        self.url = url

    def close(self):
        if self.file is not None:
            self.file.close()
            self.file = None
            self.closed = True

    def __getattr__(self, name):
        return getattr(self.file, name)

    def __enter__(self):
        return self

    def __exit__(self, t, v, tb):
        self.close()



##############################################################################
#
# Copyright (c) 2018 Zope Foundation and Contributors.
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################
"""\
Tests for ZConfig.components.logger.formatter.

"""

import logging
import sys
import unittest

import ZConfig.components.logger.formatter
import ZConfig.components.logger.tests.support


# In Python 3.8, a KeyError raised by string interpolation is re-written
# into a ValueError reporting a reference to an undefined field.  We're
# not masking the exception, but we want to check for the right one in
# the tests below (without catching anything else).
#
if sys.version_info >= (3, 8):
    MissingFieldError = ValueError
else:
    MissingFieldError = KeyError


class LogFormatStyleTestCase(unittest.TestCase):

    def setUp(self):
        unittest.TestCase.setUp(self)
        self.convert = ZConfig.components.logger.formatter.log_format_style

    def test_classic(self):
        self.assertEqual(self.convert('classic'), 'classic')
        self.assertEqual(self.convert('Classic'), 'classic')
        self.assertEqual(self.convert('CLASSIC'), 'classic')
        self.assertEqual(self.convert('cLaSsIc'), 'classic')

    def test_format(self):
        self.assertEqual(self.convert('format'), 'format')
        self.assertEqual(self.convert('Format'), 'format')
        self.assertEqual(self.convert('FORMAT'), 'format')
        self.assertEqual(self.convert('fOrMaT'), 'format')

    def test_template(self):
        self.assertEqual(self.convert('template'), 'template')
        self.assertEqual(self.convert('Template'), 'template')
        self.assertEqual(self.convert('TEMPLATE'), 'template')
        self.assertEqual(self.convert('tEmPlAtE'), 'template')

    def test_safe_template(self):
        self.assertEqual(self.convert('safe-template'), 'safe-template')
        self.assertEqual(self.convert('Safe-Template'), 'safe-template')
        self.assertEqual(self.convert('SAFE-TEMPLATE'), 'safe-template')
        self.assertEqual(self.convert('sAfE-tEmPlAtE'), 'safe-template')

    def test_bad_values(self):

        def check(value):
            with self.assertRaises(ValueError) as cm:
                self.convert(value)
            self.assertIn('log_format_style must be one of',
                          str(cm.exception))
            self.assertIn('found %r' % value, str(cm.exception))

        check('')
        check('just-some-junk')
        check('Another Pile of Junk')
        check('%')
        check('{')
        check('$')


class StyledFormatterTestHelper(
        ZConfig.components.logger.tests.support.LoggingTestHelper):

    _schematext = """
      <schema>
        <import package='ZConfig.components.logger'/>
        <section type='eventlog' name='*' attribute='eventlog'/>
      </schema>
    """

    _config_template = """\
      <eventlog>
        <logfile>
          path STDOUT
          level debug
          %s
        </logfile>
      </eventlog>
    """

    def setUp(self):
        ZConfig.components.logger.tests.support.LoggingTestHelper.setUp(self)
        self.record = logging.LogRecord(
            'ZConfig.foo.bar', logging.WARN, __file__, 42,
            'my message, %r %r', ('with', 'some args'), None, 'faux_func')
        self.record.x1 = 24
        self.record.x2 = 37

    def get_logger_factory(self, style=None, format=None,
                           arbitrary_fields=False):
        formatter_lines = []
        if style:
            formatter_lines.append('style %s' % style)
        if format:
            formatter_lines.append('format %s' % format)
        if arbitrary_fields:
            formatter_lines.append('arbitrary-fields true')
        formatter_config = '\n    '.join(formatter_lines)
        parsed = self.get_config(self._config_template % formatter_config)
        return parsed.eventlog

    def get_formatter_factory(self, style=None, format=None,
                              arbitrary_fields=False):
        logger_factory = self.get_logger_factory(
            style=style, format=format, arbitrary_fields=arbitrary_fields)
        return logger_factory.handler_factories[0].create_formatter

    def get_formatter(self, style=None, format=None, arbitrary_fields=False):
        factory = self.get_formatter_factory(
            style=style, format=format, arbitrary_fields=arbitrary_fields)
        return factory()


class LoggerStyledFormatterTestCase(StyledFormatterTestHelper,
                                    unittest.TestCase):

    def test_classic_explicit(self):
        formatter = self.get_formatter(
            style='classic',
            format='%(levelname)s %(levelno)s %(message)s')
        content = formatter.format(self.record)
        self.assertEqual(content, "WARNING 30 my message, 'with' 'some args'")

    def test_classic_implicit(self):
        formatter = self.get_formatter(
            format='%(levelname)s %(levelno)s %(message)s')
        content = formatter.format(self.record)
        self.assertEqual(content, "WARNING 30 my message, 'with' 'some args'")

    def test_format(self):
        #
        # The last item here isn't a format string placeholder; it
        # almost looks like a classic format operation, but would cause
        # an error.  We're just expecting it to show up in the result as
        # a literal.
        #
        # It would not be allowed in older versions of ZConfig, or for
        # classic style.
        #
        formatter = self.get_formatter(
            style='format',
            format='{levelname} {levelno} {message} %(stuff)')
        content = formatter.format(self.record)
        self.assertEqual(
            content, "WARNING 30 my message, 'with' 'some args' %(stuff)")

    def test_format_with_anonymous_placeholders(self):
        with self.assertRaises(ValueError):
            self.get_formatter_factory(
                style='format',
                format='{} {} {} %(stuff)')

    def test_format_with_positional_placeholders(self):
        with self.assertRaises(ValueError):
            self.get_formatter_factory(
                style='format',
                format='{1} {2} {3} %(stuff)')

    # These comments apply to the test_template_* and
    # test_safe_template_* tests that demonstrate successful formatter
    # creations.
    #
    # The last item here isn't a format string placeholder; it almost
    # looks like a classic format operation, but would cause an error.
    # We're just expecting it to show up in the result as a literal.
    #
    # It would not be allowed in older versions of ZConfig, or for
    # classic style.
    #
    # Note the $ must be doubled to get through the ZConfig syntax for
    # substitutions.
    #
    def test_template_with_braces(self):
        formatter = self.get_formatter(
            style='template',
            format='$${levelname} $${levelno} $${message} %(stuff) {extra}')
        content = formatter.format(self.record)
        self.assertEqual(
            content, ("WARNING 30 my message, 'with' 'some args'"
                      " %(stuff) {extra}"))

    def test_template_without_braces(self):
        formatter = self.get_formatter(
            style='template',
            format='$$levelname $$levelno $$message %(stuff) {extra}')
        content = formatter.format(self.record)
        self.assertEqual(
            content, ("WARNING 30 my message, 'with' 'some args'"
                      " %(stuff) {extra}"))

    def test_template_with_junk(self):
        #
        # There's a lot of junk in this format; we're verifying that
        # it's caught when constructing the factory.
        #
        with self.assertRaises(ValueError):
            self.get_formatter_factory(
                style='format',
                format='$$} $${levelno')

    def test_safe_template_with_braces(self):
        formatter = self.get_formatter(
            style='safe-template',
            format='$${levelname} $${levelno} $${message} %(stuff) {extra}')
        content = formatter.format(self.record)
        self.assertEqual(
            content, ("WARNING 30 my message, 'with' 'some args'"
                      " %(stuff) {extra}"))

    def test_safe_template_without_braces(self):
        formatter = self.get_formatter(
            style='safe-template',
            format='$$levelname $$levelno $$message %(stuff) {extra}')
        content = formatter.format(self.record)
        self.assertEqual(
            content, ("WARNING 30 my message, 'with' 'some args'"
                      " %(stuff) {extra}"))

    def test_safe_template_with_junk(self):
        formatter = self.get_formatter(
            style='safe-template',
            format=('$${levelname} $${levelno} $${message} %(stuff) {extra}'
                    ' $$} $${levelno  $${bad-mojo}'))
        content = formatter.format(self.record)
        self.assertEqual(
            content, ("WARNING 30 my message, 'with' 'some args'"
                      " %(stuff) {extra} $} ${levelno  ${bad-mojo}"))


class ZopeExceptionsFormatterTestCase(LoggerStyledFormatterTestCase):

    # We test against the zope.exceptions formatter since it's a common
    # example of a formatter that inherits almost everything, but
    # defines it's own formatException.

    _config_template = """\
      <eventlog>
        <logfile>
          path STDOUT
          level debug
          formatter zope.exceptions.log.Formatter
          %s
        </logfile>
      </eventlog>
    """

    def test_format_with_traceback_info(self):
        formatter = self.get_formatter(
            style='format',
            format='{levelname} {levelno} {message}')

        def fail():
            raise RuntimeError('foo')

        def something():
            __traceback_info__ = 42
            fail()

        try:
            something()
        except RuntimeError:
            self.record.exc_info = sys.exc_info()

        content = formatter.format(self.record)
        self.assertIn(' - __traceback_info__: 42', content)
        self.assertIn("WARNING 30 my message, 'with' 'some args'", content)


class CustomFormatterClassWithoutStyleParamTestCase(
        LoggerStyledFormatterTestCase):

    _config_template = """\
      <eventlog>
        <logfile>
          path STDOUT
          level debug
          formatter %s.StylelessFormatter
          %%s
        </logfile>
      </eventlog>
    """ % __name__


class CustomFormatterFactoryWithoutStyleParamTestCase(
        LoggerStyledFormatterTestCase):

    _config_template = """\
      <eventlog>
        <logfile>
          path STDOUT
          level debug
          formatter %s.styleless_formatter
          %%s
        </logfile>
      </eventlog>
    """ % __name__


class StylelessFormatter(logging.Formatter):

    def __init__(self, fmt=None, datefmt=None):
        kwargs = dict()
        if sys.version_info >= (3, 8):
            kwargs['validate'] = False
        logging.Formatter.__init__(self, fmt=fmt, datefmt=datefmt, **kwargs)


def styleless_formatter(fmt=None, datefmt=None):
    return StylelessFormatter(fmt=fmt, datefmt=datefmt)


class FieldTypesTestCase(StyledFormatterTestHelper, unittest.TestCase):

    def test_func_name_classic(self):
        formatter = self.get_formatter(
            style='classic',
            format='%(levelname)s %(levelno)2d %(funcName)s')
        output = formatter.format(self.record)
        self.assertIn('WARNING 30 faux_func', output)

    def test_func_name_format(self):
        formatter = self.get_formatter(
            style='format',
            format='{levelname} {levelno:02d} {funcName}')
        output = formatter.format(self.record)
        self.assertIn('WARNING 30 faux_func', output)

    def test_func_name_template(self):
        formatter = self.get_formatter(
            style='template',
            format='$$levelname $$levelno $$funcName')
        output = formatter.format(self.record)
        self.assertIn('WARNING 30 faux_func', output)

    def test_levelno_integer_classic(self):
        formatter = self.get_formatter(
            style='classic',
            format='%(levelname)s %(levelno)2d %(message)s')
        output = formatter.format(self.record)
        self.assertIn('WARNING 30 my message', output)

    def test_levelno_integer_format(self):
        formatter = self.get_formatter(
            style='format',
            format='{levelname} {levelno:02d} {message}')
        output = formatter.format(self.record)
        self.assertIn('WARNING 30 my message', output)

    def test_msecs_float_classic(self):
        formatter = self.get_formatter(
            style='classic',
            format='%(asctime)s.%(msecs)03.0f %(levelname)s %(message)s')
        self.record.msecs = 619.041919708252
        output = formatter.format(self.record)
        expected = '%s.619 WARNING my message' % self.record.asctime
        self.assertIn(expected, output)

    def test_msecs_float_format(self):
        formatter = self.get_formatter(
            style='format',
            format='{asctime}.{msecs:03.0f} {levelname} {message}')
        self.record.msecs = 619.041919708252
        output = formatter.format(self.record)
        expected = '%s.619 WARNING my message' % self.record.asctime
        self.assertIn(expected, output)

    def test_relative_created_float_classic(self):
        formatter = self.get_formatter(
            style='classic',
            format='%(relativeCreated)+.3f %(levelname)s %(message)s')
        self.record.relativeCreated = 406.7840576171875
        output = formatter.format(self.record)
        self.assertIn('+406.784 WARNING my message', output)

    def test_relative_created_float_format(self):
        formatter = self.get_formatter(
            style='format',
            format='{relativeCreated:+.3f} {levelname} {message}')
        self.record.relativeCreated = 406.7840576171875
        output = formatter.format(self.record)
        self.assertIn('+406.784 WARNING my message', output)


class ArbitraryFieldsTestCase(StyledFormatterTestHelper, unittest.TestCase):

    def test_classic_arbitrary_field_disallowed_by_default(self):
        with self.assertRaises(KeyError) as cm:
            self.get_formatter_factory(
                style='classic',
                format='%(levelno)s %(levelname)s %(undefined_field)s')
        self.assertEqual(str(cm.exception), "'undefined_field'")

    def test_format_arbitrary_field_disallowed_by_default(self):
        with self.assertRaises(KeyError) as cm:
            self.get_formatter_factory(
                style='format',
                format='{levelno} {levelname} {undefined_field}')
        self.assertEqual(str(cm.exception), "'undefined_field'")

    def test_template_arbitrary_field_disallowed_by_default(self):
        with self.assertRaises(KeyError) as cm:
            self.get_formatter_factory(
                style='format',
                format='$${levelno} $${levelname} $${undefined_field}')
        self.assertEqual(str(cm.exception), "'undefined_field'")

    # We don't need to check the safe-template variety, because it's
    # highly permissive.

    def test_classic_arbitrary_field_missing(self):
        formatter = self.get_formatter(
            style='classic',
            format='%(levelno)s %(levelname)s %(undefined_field)s',
            arbitrary_fields=True)

        # The formatter still breaks when it references an undefined field:
        with self.assertRaises(MissingFieldError) as cm:
            formatter.format(self.record)
        self.assertIn("'undefined_field'", str(cm.exception))

    def test_classic_arbitrary_field_present(self):
        formatter = self.get_formatter(
            style='classic',
            format='%(levelno)s %(levelname)s %(undefined_field)s',
            arbitrary_fields=True)

        # Given the field, it formats just fine:
        self.record.undefined_field = 'surprise!'
        logentry = formatter.format(self.record)
        self.assertIn('surprise!', logentry)

    def test_format_arbitrary_field_missing(self):
        formatter = self.get_formatter(
            style='format',
            format='{levelno} {levelname} {undefined_field}',
            arbitrary_fields=True)

        # The formatter still breaks when it references an undefined field:
        with self.assertRaises(MissingFieldError) as cm:
            formatter.format(self.record)
        self.assertIn("'undefined_field'", str(cm.exception))

    def test_format_arbitrary_field_present(self):
        formatter = self.get_formatter(
            style='format',
            format='{levelno} {levelname} {undefined_field}',
            arbitrary_fields=True)

        # Given the field, it formats just fine:
        self.record.undefined_field = 'surprise!'
        logentry = formatter.format(self.record)
        self.assertIn('surprise!', logentry)

    def test_template_arbitrary_field_missing(self):
        formatter = self.get_formatter(
            style='template',
            format='$${levelno} $${levelname} $${undefined_field}',
            arbitrary_fields=True)

        # The formatter still breaks when it references an undefined field:
        with self.assertRaises(MissingFieldError) as cm:
            formatter.format(self.record)
        self.assertIn("'undefined_field'", str(cm.exception))

    def test_template_arbitrary_field_present(self):
        formatter = self.get_formatter(
            style='template',
            format='$${levelno} $${levelname} $${undefined_field}',
            arbitrary_fields=True)

        # Given the field, it formats just fine:
        self.record.undefined_field = 'surprise!'
        logentry = formatter.format(self.record)
        self.assertIn('surprise!', logentry)

    def test_safe_template_arbitrary_field_missing(self):
        formatter = self.get_formatter(
            style='safe-template',
            format='$${levelno} $${levelname} $${undefined_field}',
            arbitrary_fields=True)

        # The formatter still breaks when it references an undefined field:
        logentry = formatter.format(self.record)
        self.assertIn(' ${undefined_field}', logentry)

    def test_safe_template_arbitrary_field_present(self):
        formatter = self.get_formatter(
            style='safe-template',
            format='$${levelno} $${levelname} $${undefined_field}',
            arbitrary_fields=True)

        # Given the field, it formats just fine:
        self.record.undefined_field = 'surprise!'
        logentry = formatter.format(self.record)
        self.assertIn('surprise!', logentry)
